import json

from helper import json_serialize
from messagebus import Message, DefaultToken, DataType
from messageprocessorutil import UuidMessageProcessor
from typing import Dict, Any

from customer_services.interactor.add_customer import AddCustomerInteractor
from customer_services.interactor.add_customer.dto import AddCustomerRequest

from externals.util import LoggerUtil

logger = LoggerUtil.get_logger_name(__name__)


class AddCustomerProcessor(UuidMessageProcessor):
    def __init__(self, interactor):
        # type: (AddCustomerInteractor) -> None
    
        super(AddCustomerProcessor, self).__init__()
        
        self.interactor = interactor

    def parse_data(self, data):
        # type: (Any) -> AddCustomerRequest

        request = json.loads(data.decode("utf-8"))  # type: Dict[str, str]
        return AddCustomerRequest(request["phone"],
                                  request["document"],
                                  request["name"],
                                  request["zip_code"],
                                  request["city"],
                                  request["street"],
                                  request["neighborhood"],
                                  request["number"],
                                  request["complement"],
                                  request["reference_point"])

    def call_business(self, request):
        # type: (AddCustomerRequest) -> None

        logger.info("Adding customer for request: {}".format(json_serialize(request)))
        self.interactor.add_customer(request)

    def format_response(self, message_bus, message, event, request, response):
        logger.info("Customer added for request: {}".format(json_serialize(request)))
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_ACK, DataType.param, response))

    def format_exception(self, message_bus, message, event, request, exception):
        exception_name = str(type(exception).__name__)
        logger.info("Error adding customer from request {}: {}".format(json_serialize(request), exception_name))
        reply_message = Message(DefaultToken.TK_SYS_NAK, DataType.param, exception_name)
        message_bus.reply_message(message, reply_message)

    def format_parse_exception(self, message_bus, message, event, data, exception):
        logger.info("Error getting parsing request: {}".format(str(exception)))
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))



# -*- coding: utf-8 -*-

"""
Twisted-specific convenience helpers.
"""

from __future__ import absolute_import, division, print_function

from OpenSSL.SSL import FILETYPE_PEM
from twisted.internet import ssl

from ._core import Certificate, DHParameters, Key, parse_file


# mypy hack: Import typing information without actually importing anything.
MYPY = False
if MYPY:  # pragma: nocover
    from ._core import AbstractPEMObject
    from typing import List, Any


def certificateOptionsFromPEMs(pemObjects, **kw):
    # type: (List[AbstractPEMObject], **Any) -> ssl.CerticateOptions
    """
    Load a CertificateOptions from the given collection of PEM objects
    (already-loaded private keys and certificates).

    In those PEM objects, identify one private key and its corresponding
    certificate to use as the primary certificate.  Then use the rest of the
    certificates found as chain certificates.  Raise a ValueError if no
    certificate matching a private key is found.

    :return: A TLS context factory using *pemObjects*
    :rtype: `twisted.internet.ssl.CertificateOptions`_

    .. _`twisted.internet.ssl.CertificateOptions`: \
        https://twistedmatrix.com/documents/current/api/\
        twisted.internet.ssl.CertificateOptions.html
    """
    keys = [key for key in pemObjects if isinstance(key, Key)]
    if not len(keys):
        raise ValueError("Supplied PEM file(s) does *not* contain a key.")
    if len(keys) > 1:
        raise ValueError("Supplied PEM file(s) contains *more* than one key.")

    privateKey = ssl.KeyPair.load(str(keys[0]), FILETYPE_PEM)

    certs = [cert for cert in pemObjects if isinstance(cert, Certificate)]
    if not len(certs):
        raise ValueError("*At least one* certificate is required.")
    certificates = [ssl.Certificate.loadPEM(str(certPEM)) for certPEM in certs]

    certificatesByFingerprint = dict(
        [
            (certificate.getPublicKey().keyHash(), certificate)
            for certificate in certificates
        ]
    )

    if privateKey.keyHash() not in certificatesByFingerprint:
        raise ValueError(
            "No certificate matching {fingerprint} found.".format(
                fingerprint=privateKey.keyHash()
            )
        )

    primaryCertificate = certificatesByFingerprint.pop(privateKey.keyHash())

    if "dhParameters" in kw:
        raise TypeError(
            "Passing DH parameters as a keyword argument instead of a "
            "PEM object is not supported anymore."
        )

    dhparams = [o for o in pemObjects if isinstance(o, DHParameters)]
    if len(dhparams) > 1:
        raise ValueError(
            "Supplied PEM file(s) contain(s) *more* than one set of DH "
            "parameters."
        )
    elif len(dhparams) == 1:
        kw["dhParameters"] = ssl.DiffieHellmanParameters(str(dhparams[0]))

    ctxFactory = ssl.CertificateOptions(
        privateKey=privateKey.original,
        certificate=primaryCertificate.original,
        extraCertChain=[
            chain.original for chain in certificatesByFingerprint.values()
        ],
        **kw
    )

    return ctxFactory


def certificateOptionsFromFiles(*pemFiles, **kw):
    # type: (*str, **Any) -> ssl.CertificateOptions
    """
    Read all files named by *pemFiles*, and parse them using
    :func:`certificateOptionsFromPEMs`.
    """
    pems = []  # type: List[AbstractPEMObject]
    for pemFile in pemFiles:
        pems += parse_file(pemFile)

    return certificateOptionsFromPEMs(pems, **kw)



import json
import logging

from xml.etree import cElementTree as eTree

from model.customexception import ErrorApplyingDiscount, DiscountMinLevelException
from typing import Optional

from actions.sale_line.set_custom_properties import set_custom_properties
from bustoken import TK_NCOUPON_APPLY_COUPON
from msgbus import FM_PARAM, TK_SYS_ACK, MBTimeout, MBException
from mw_helper import (
    show_keyboard,
)
from helper import remove_accents
from sysactions import action, get_model, get_current_order, show_messagebox, get_posot, translate_message

from actions import mb_context
from custom_sysactions import user_authorized_to_execute_action

logger = logging.getLogger("PosActions")


@action
@user_authorized_to_execute_action
def do_apply_coupons(pos_id, table_id=None, seat=''):
    coupon_code = show_keyboard(pos_id, "$ENTER_COUPON_CODE", defvalue="", title="", numpad=False)
    if coupon_code is None:
        return False

    if seat == '':
        apply_coupon(pos_id, table_id=table_id, coupon_code=coupon_code)
        return

    _apply_coupon_and_set_seat(coupon_code, pos_id, seat, table_id)


def apply_coupon(
    pos_id,                 # type: str
    table_id=None,          # type: Optional[str]
    coupon_code=None,       # type: Optional[str]
    user_auth_level=-1,     # type: Optional[int]
    raise_error=False       # type: Optional[bool]
):
    # type: (...) -> None

    model = get_model(pos_id)
    order = get_current_order(model)
    message = None

    try:
        coupon_code = coupon_code.upper()
        originator_pos_id = int(order.get("originatorId")[3:]) if order else None
        data = json.dumps(dict(
            posId=pos_id,
            originatorPosId=originator_pos_id,
            voucherId=coupon_code,
            user_auth_level=user_auth_level
        ))

        response = mb_context.MB_EasySendMessage("nCoupon", TK_NCOUPON_APPLY_COUPON, FM_PARAM, data)
        if response.token != TK_SYS_ACK:
            if response.data == "NDISCOUNT_ERROR_USER_MIN_LEVEL":
                raise DiscountMinLevelException()

            exception = ErrorApplyingDiscount()
            exception.message = _format_error_message(
                model=model,
                data=response.data,
            )
            raise exception

        message = "$SUCCESS_NOT_TOTALED_COUPON_APPLIED"
        if response.token == TK_SYS_ACK and order and order.get('state') == "TOTALED":
            message = "$SUCCESS_COUPON_APPLIED"
            if table_id is not None:
                from table_actions import process_discount_in_table
                process_discount_in_table(pos_id, order, table_id)

    except (MBTimeout, MBException) as _:
        show_messagebox(pos_id, "$ERROR_COUPON_APPLIED")
        logger.exception("Error sending message to nDiscount")
        raise
    except DiscountMinLevelException as _:
        logger.exception("Error applying coupon")
        raise
    except Exception as ex:
        error_message = _get_error_message_from_exception(
            exception=ex,
            model=model,
        )

        show_messagebox(pos_id, error_message)
        logger.exception("Error applying coupon")
        if raise_error:
            raise
    finally:
        if message:
            show_messagebox(pos_id, message)


def _get_error_message_from_exception(
    exception,  # type: Exception
    model,      # type: eTree.XML
):
    # type: (...) -> str

    error_message = None
    if hasattr(exception, 'message') and exception.message not in (None, "None", ""):
        error_message = translate_message(
            model=model,
            message_key=str(exception.message),
        )

    return error_message or translate_message(
        model=model,
        message_key="ERROR_COUPON_APPLIED",
    )


def _apply_coupon_and_set_seat(coupon_code, pos_id, seat, table_id):
    # type: (str, str, str, str) -> None

    model = get_model(pos_id)
    pos_ot = get_posot(model)
    order_picture = eTree.XML(pos_ot.orderPicture(pos_id))
    order_id = order_picture.get("orderId")
    sale_lines = order_picture.findall(".//SaleLine")
    last_line_number = max([int(x.get("lineNumber")) for x in sale_lines]) if sale_lines else 0
    apply_coupon(pos_id, table_id=table_id, coupon_code=coupon_code)
    _set_custom_properties_for_added_sale_lines(pos_id, pos_ot, order_id, last_line_number, seat)


def _set_custom_properties_for_added_sale_lines(pos_id, pos_ot, order_id, last_line_number, seat):
    order_picture = eTree.XML(pos_ot.orderPicture(pos_id, order_id))
    sale_lines = order_picture.findall(".//SaleLine")
    last_line_number_after_apply_coupon = max([int(x.get("lineNumber")) for x in sale_lines]) if sale_lines else 0
    if last_line_number != last_line_number_after_apply_coupon:
        sale_type = None if order_picture.get("saleTypeDescr") == "EAT_IN" else order_picture.get("saleTypeDescr")

        for sale_line in sale_lines:
            if int(sale_line.get("level")) > 0 or int(sale_line.get("lineNumber")) <= int(last_line_number):
                continue

            set_custom_properties(pos_ot, sale_line, seat, sale_type)


def _format_error_message(
    model,    # type: eTree.Element
    data,     # type: json
):
    # type: (...) -> Optional[str]

    if data in (None, "None", ""):
        return

    try:
        data = json.loads(data)

        if data.get("status_code") != 406:
            return data.get("message") or None

        language = model.findall("Language")[0].get("name").lower().replace("_", "-")
        data_message = data.get("message", {}).get("messages", "") if data.get("message") else {}
        translated_message = data_message.get(language, "")
        return remove_accents(text=translated_message)
    except Exception as _:
        return data




# -*- coding: utf-8 -*-
import inspect
from functools import wraps

from sysactions import get_model, StopAction, show_messagebox, get_podtype, get_line_product_name
from systools import sys_log_exception

from actions.config import get_function_authorization_level_from_config
from actions.util import get_user_level, get_authorization
from actions.pos import pos_need_to_be_blocked

configurations = None


def initialize_manager_authorizations(config):
    global configurations
    configurations = config


def user_authorized_to_execute_action(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        user_data = None

        pos_id = args[0]
        try:
            function_level = get_function_authorization_level_from_config(configurations, f.func_name)
            model = get_model(pos_id)
            operator = model.find("./Operator")
            if operator is not None:
                user_id = operator.get("id")
                if get_user_level(user_id) <= function_level:
                    status, user_data = get_authorization(pos_id, function_level, model, executed_function=f.func_name)
                    if not status:
                        return StopAction()
        except BaseException as ex: # noqa
            error_message = "$GET_USER_AUTHORIZATION_ERROR_" + str(type(ex).__name__).upper()
            sys_log_exception("Error validating user authorization")
            show_messagebox(pos_id, message=error_message, icon="error")
            return False

        if 'auth_data' in inspect.getargspec(f).args:
            kwargs['auth_data'] = user_data

        return f(*args, **kwargs)

    return decorated_function


def block_action_if_pos_is_blocked(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        try:
            pos_id = args[0]
            model = get_model(pos_id)
            pod_type = get_podtype(model)
            if pod_type != 'TT' and pos_need_to_be_blocked(pos_id):
                from msgbusboundary import MsgBusTableService as tableService
                from table_actions import deselect_table
                selected_table = tableService().get_selected_table(pos_id)
                if selected_table is None:
                    deselect_table(pos_id)

                show_messagebox(pos_id, message="$POS_BLOCKED", icon="error")
                return StopAction()
        except BaseException as _: # noqa
            return False

        return f(*args, **kwargs)

    return decorated_function


def block_action_if_order_items_not_resolved(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        from posactions import list_open_options

        pos_id = args[0]
        model = get_model(pos_id)
        order = model.find("CurrentOrder/Order")

        sale_lines = order.findall("SaleLine")
        active_sale_lines = _get_active_order_lines(sale_lines)
        is_remote_order = order.find("CustomOrderProperties/OrderProperty[@key='REMOTE_ORDER_ID']")
        if not active_sale_lines and is_remote_order is None:
            show_messagebox(pos_id, "$HAS_NO_ITEMS_IN_ORDER", title="")
            return False

        option = list_open_options(order)
        if option is not None:
            prod_name = get_line_product_name(model, int(option.get("lineNumber")))
            show_messagebox(pos_id, message="$NEED_TO_RESOLVE_OPTION|%s" % (prod_name.encode("UTF-8")), icon="error")
            return False

        return f(*args, **kwargs)

    return decorated_function


def _get_order_deleted_lines(sale_lines):
    return map(lambda x: x.get("lineNumber"), filter(lambda x: x.get("level") == "0" and x.get("qty") == "0", sale_lines))


def _get_active_order_lines(sale_lines):
    deleted_line_numbers = _get_order_deleted_lines(sale_lines)

    return filter(lambda x: x.get("lineNumber") not in deleted_line_numbers, sale_lines)



from ._BumpBar import BumpBar
from ._Kds import Kds
from ._KdsView import KdsView
from ._ZoomLevel import ZoomLevel
from ._Modifier import Modifier
from ._BumpBarCommand import BumpBarCommand
from ._KdsModel import KdsModel
from ._AutoOrderCommand import AutoOrderCommand



from _CashReportBodyDto import CashReportBodyDto  # noqa


class CloseDaySummaryBodyDto(object):
    def __init__(self, authorizer, open_day_time, report_body):
        # type: (unicode, unicode, CashReportBodyDto) -> ()
        self.authorizer = authorizer
        self.open_day_time = open_day_time
        self.report_body = report_body



from application.domain.ui import ListBox
from sysactions import show_listbox


class MsgBusListBox(ListBox):
    def show_listbox(self, pos_id, options, message='$SELECT_AN_OPTION', title='', def_value='', buttons='$OK|$CANCEL',
                     icon='', timeout=60000, asynch=False):
        selected = show_listbox(pos_id, options, "$SELECT_SEAT_FOR_LINE", "$SEAT_SELECTION_TITLE", def_value, buttons,
                                icon, timeout, asynch)
        return selected



from abc import (
    ABCMeta,
    abstractmethod,
)
from datetime import datetime
from typing import List


class OrderRepository(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_paid_ticket(
        self,
        business_period,    # type: str
    ):
        # type: (...) -> dict

        raise NotImplementedError

    @abstractmethod
    def get_paid_orders_ticket(
        self,
        business_period,    # type: str
    ):
        # type: (...) -> List

        raise NotImplementedError

    @abstractmethod
    def get_paid_orders(
        self,
        operator_id,    # type: str
        initial_date,   # type: datetime
        end_date,       # type: datetime
    ):
        # type: (...) -> List

        raise NotImplementedError

    @abstractmethod
    def get_orders_items(
        self,
        paid_orders,    # type: List
    ):
        # type: (...) -> List

        raise NotImplementedError



from model.loyalty import (
    AdditionalInfo,
    Customer,
)


class CustomerInfo(object):
    def __init__(
        self,
        loyalty,    # type: Dict
    ):
        # type: (...) -> None

        customer_info = loyalty.get("customerInfo", {})

        customer = customer_info.get("customer", {})
        if customer:
            self.customer = Customer(
                customer=customer,
            )

        additional_info = customer_info.get("additionalInfo", {})
        if additional_info:
            self.additional_info = AdditionalInfo(
                additional_info=additional_info,
            )



from sysactions import show_messagebox, show_listbox, get_model, print_report, action, show_keyboard, StopAction

from application.repository import AccountRepository
from actions.util import get_operator_info
from manager.reports import get_report_date, is_valid_date_range
from typing import Tuple

from .. import mb_context


@action
def show_operator_closing(pos_id, operator_id):
    end_date, operator_id, start_date = get_manual_inserted_data(operator_id, pos_id)

    model = get_model(pos_id)
    _, _, session = get_operator_info(model, operator_id)

    options, sessions = get_date_options_and_sessions(end_date, operator_id, pos_id, session, start_date)

    selected_date = show_listbox(pos_id, options, message="$SELECT_A_DATE")
    if selected_date is None:
        raise StopAction()

    session = sessions[selected_date]
    selected_date = session.split("period=")[1]

    print_report(pos_id, model, True, "close_operator_report", pos_id, selected_date, operator_id, session)


def get_manual_inserted_data(
        operator_id,    # type: str
        pos_id          # type: str
):
    # type: (...) -> Tuple[str, str, str]

    start_date = get_report_date(pos_id, "$TYPE_INITIAL_DATE")
    if start_date is None:
        raise StopAction()

    end_date = get_report_date(pos_id, "$TYPE_FINAL_DATE")
    if end_date is None:
        raise StopAction()

    if is_valid_date_range(start_date, end_date, pos_id) is False:
        raise StopAction()

    if not operator_id:
        operator_id = show_keyboard(pos_id, "$TYPE_OPERATOR_ID_MESSAGE", mask="INTEGER", numpad=True)
        if operator_id is None:
            raise StopAction()

    return end_date, operator_id, start_date


def get_date_options_and_sessions(end_date, operator_id, pos_id, session, start_date):
    account_repository = AccountRepository(mb_context)
    sessions = account_repository.get_closing_operator_sessions(pos_id, operator_id, start_date, end_date)

    if session in sessions:
        sessions.remove(session)

    if not sessions:
        show_messagebox(pos_id, "$THERE_ARE_NO_DATA_TO_SHOW", "$INFORMATION")
        raise StopAction()

    options = []
    for current_session in sessions:
        session = current_session.split("period=")[1].split(",")[0]
        counter = current_session.split("count=")[1].split(",")[0]
        pos_type = 'QSR' if 'pos' in current_session else 'TS'
        formatted_option = "{}/{}/{} {} ({})".format(session[6:], session[4:6], session[:4], pos_type, counter)
        options.append(formatted_option)

    return options, sessions



# -*- coding: utf-8 -*-

from sysactions import action, show_messagebox
from bustoken import TK_REMOTE_ORDER_SEARCH_LOGISTIC
from msgbus import TK_SYS_ACK

from .. import mb_context
from model.customexception import LogisticRequestError


@action
def request_logistic_remote_order(pos_id, order_id=""):
    data = pos_id + "\0" + order_id
    search_logistic_token = TK_REMOTE_ORDER_SEARCH_LOGISTIC
    msg = mb_context.MB_EasySendMessage("RemoteOrder", search_logistic_token, data=data, timeout=10000 * 1000)
    if msg.token == TK_SYS_ACK:
        show_messagebox(pos_id, "$LOGISTIC_SEARCHING")
        return
    
    raise LogisticRequestError("$LOGISTIC_SEARCHING_ERROR")



from logging import Logger
from typing import Optional

from production.box._OrderChangerProductionBox import OrderChangerProductionBox


class CFHHiderBox(OrderChangerProductionBox):
    def __init__(
        self,
        name,               # type: str
        sons,               # type: object
        cfh_items,          # type: object
        inherit_comments,   # type: bool
        logger=None,        # type: Optional[Logger]
    ):
        # type: (...) -> None

        super(CFHHiderBox, self).__init__(name, sons, logger)
        self.cfh_items = cfh_items
        self.inherit_comments = inherit_comments

    def change_order(self, order):
        order.items = self._find_and_remove_cfh_items(order.items[:])
        return order

    def _find_and_remove_cfh_items(self, items):
        new_items = []
        for item in items:
            if item.part_code in self.cfh_items:
                for son in item.items:
                    son.handle_remove_father(
                        father=item,
                        inherit_comments=self.inherit_comments,
                    )
                    new_items.extend(self._find_and_remove_cfh_items([son]))
            else:
                for son in item.items:
                    if self.inherit_comments and son.part_code in self.cfh_items:
                        son.repeat_comment_from_father(father=item)
                item.items = self._find_and_remove_cfh_items(item.items)
                new_items.append(item)

        return new_items



from report.command import NewLineCommand

from .._CommandProcessor import CommandProcessor


class NewLineCommandProcessor(CommandProcessor):
    def get_bytes(self, command):
        if not isinstance(command, NewLineCommand):
            return ""

        return "\n"



from abc import ABCMeta, abstractmethod
from datetime import datetime

from production.model import ProductionOrder
from typing import (
    List,
    Dict,
)


class ProductionRepository(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def save_order(self, order):
        # type: (ProductionOrder) -> None

        raise NotImplementedError()

    def get_all_orders(self):
        # type: () -> List[ProductionOrder]

        raise NotImplementedError()

    def get_orders_to_purge(self, limit_date):
        # type: (datetime) -> List[ProductionOrder]

        raise NotImplementedError()

    def purge_orders(self, orders):
        # type: (List[ProductionOrder]) -> None

        raise NotImplementedError()

    def stop(self):
        # type: () -> None

        raise NotImplementedError()

    @abstractmethod
    def save_pick_list_order(
        self,
        order_id,           # type: int
        order_id_count,     # type: str
        jit,                # type: str
        course,             # type: str
        receipt,            # type: str
        status,             # type: bool
    ):
        # type: (...) -> None

        raise NotImplementedError()

    def get_all_pick_list_orders(
        self,
        orders_list,  # type: List[str]
    ):
        # type: (...) -> Dict

        raise NotImplementedError()

    def get_data_pick_list_order(
        self,
        pick_list_order_id,  # type: int
    ):
        # type: (...) -> str

        raise NotImplementedError()



#! python
#
# This module implements a loop back connection receiving itself what it sent.
#
# The purpose of this module is.. well... You can run the unit tests with it.
# and it was so easy to implement ;-)
#
# This file is part of pySerial. https://github.com/pyserial/pyserial
# (C) 2001-2020 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause
#
# URL format:    loop://[option[/option...]]
# options:
# - "debug" print diagnostic messages
from __future__ import absolute_import

import logging
import numbers
import time
try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse
try:
    import queue
except ImportError:
    import Queue as queue

from serial.serialutil import SerialBase, SerialException, to_bytes, iterbytes, SerialTimeoutException, PortNotOpenError

# map log level names to constants. used in from_url()
LOGGER_LEVELS = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
}


class Serial(SerialBase):
    """Serial port implementation that simulates a loop back connection in plain software."""

    BAUDRATES = (50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800,
                 9600, 19200, 38400, 57600, 115200)

    def __init__(self, *args, **kwargs):
        self.buffer_size = 4096
        self.queue = None
        self.logger = None
        self._cancel_write = False
        super(Serial, self).__init__(*args, **kwargs)

    def open(self):
        """\
        Open port with current settings. This may throw a SerialException
        if the port cannot be opened.
        """
        if self.is_open:
            raise SerialException("Port is already open.")
        self.logger = None
        self.queue = queue.Queue(self.buffer_size)

        if self._port is None:
            raise SerialException("Port must be configured before it can be used.")
        # not that there is anything to open, but the function applies the
        # options found in the URL
        self.from_url(self.port)

        # not that there anything to configure...
        self._reconfigure_port()
        # all things set up get, now a clean start
        self.is_open = True
        if not self._dsrdtr:
            self._update_dtr_state()
        if not self._rtscts:
            self._update_rts_state()
        self.reset_input_buffer()
        self.reset_output_buffer()

    def close(self):
        if self.is_open:
            self.is_open = False
            try:
                self.queue.put_nowait(None)
            except queue.Full:
                pass
        super(Serial, self).close()

    def _reconfigure_port(self):
        """\
        Set communication parameters on opened port. For the loop://
        protocol all settings are ignored!
        """
        # not that's it of any real use, but it helps in the unit tests
        if not isinstance(self._baudrate, numbers.Integral) or not 0 < self._baudrate < 2 ** 32:
            raise ValueError("invalid baudrate: {!r}".format(self._baudrate))
        if self.logger:
            self.logger.info('_reconfigure_port()')

    def from_url(self, url):
        """extract host and port from an URL string"""
        parts = urlparse.urlsplit(url)
        if parts.scheme != "loop":
            raise SerialException(
                'expected a string in the form '
                '"loop://[?logging={debug|info|warning|error}]": not starting '
                'with loop:// ({!r})'.format(parts.scheme))
        try:
            # process options now, directly altering self
            for option, values in urlparse.parse_qs(parts.query, True).items():
                if option == 'logging':
                    logging.basicConfig()   # XXX is that good to call it here?
                    self.logger = logging.getLogger('pySerial.loop')
                    self.logger.setLevel(LOGGER_LEVELS[values[0]])
                    self.logger.debug('enabled logging')
                else:
                    raise ValueError('unknown option: {!r}'.format(option))
        except ValueError as e:
            raise SerialException(
                'expected a string in the form '
                '"loop://[?logging={debug|info|warning|error}]": {}'.format(e))

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    @property
    def in_waiting(self):
        """Return the number of bytes currently in the input buffer."""
        if not self.is_open:
            raise PortNotOpenError()
        if self.logger:
            # attention the logged value can differ from return value in
            # threaded environments...
            self.logger.debug('in_waiting -> {:d}'.format(self.queue.qsize()))
        return self.queue.qsize()

    def read(self, size=1):
        """\
        Read size bytes from the serial port. If a timeout is set it may
        return less characters as requested. With no timeout it will block
        until the requested number of bytes is read.
        """
        if not self.is_open:
            raise PortNotOpenError()
        if self._timeout is not None and self._timeout != 0:
            timeout = time.time() + self._timeout
        else:
            timeout = None
        data = bytearray()
        while size > 0 and self.is_open:
            try:
                b = self.queue.get(timeout=self._timeout)  # XXX inter char timeout
            except queue.Empty:
                if self._timeout == 0:
                    break
            else:
                if b is not None:
                    data += b
                    size -= 1
                else:
                    break
            # check for timeout now, after data has been read.
            # useful for timeout = 0 (non blocking) read
            if timeout and time.time() > timeout:
                if self.logger:
                    self.logger.info('read timeout')
                break
        return bytes(data)

    def cancel_read(self):
        self.queue.put_nowait(None)

    def cancel_write(self):
        self._cancel_write = True

    def write(self, data):
        """\
        Output the given byte string over the serial port. Can block if the
        connection is blocked. May raise SerialException if the connection is
        closed.
        """
        self._cancel_write = False
        if not self.is_open:
            raise PortNotOpenError()
        data = to_bytes(data)
        # calculate aprox time that would be used to send the data
        time_used_to_send = 10.0 * len(data) / self._baudrate
        # when a write timeout is configured check if we would be successful
        # (not sending anything, not even the part that would have time)
        if self._write_timeout is not None and time_used_to_send > self._write_timeout:
            # must wait so that unit test succeeds
            time_left = self._write_timeout
            while time_left > 0 and not self._cancel_write:
                time.sleep(min(time_left, 0.5))
                time_left -= 0.5
            if self._cancel_write:
                return 0  # XXX
            raise SerialTimeoutException('Write timeout')
        for byte in iterbytes(data):
            self.queue.put(byte, timeout=self._write_timeout)
        return len(data)

    def reset_input_buffer(self):
        """Clear input buffer, discarding all that is in the buffer."""
        if not self.is_open:
            raise PortNotOpenError()
        if self.logger:
            self.logger.info('reset_input_buffer()')
        try:
            while self.queue.qsize():
                self.queue.get_nowait()
        except queue.Empty:
            pass

    def reset_output_buffer(self):
        """\
        Clear output buffer, aborting the current output and
        discarding all that is in the buffer.
        """
        if not self.is_open:
            raise PortNotOpenError()
        if self.logger:
            self.logger.info('reset_output_buffer()')
        try:
            while self.queue.qsize():
                self.queue.get_nowait()
        except queue.Empty:
            pass

    @property
    def out_waiting(self):
        """Return how many bytes the in the outgoing buffer"""
        if not self.is_open:
            raise PortNotOpenError()
        if self.logger:
            # attention the logged value can differ from return value in
            # threaded environments...
            self.logger.debug('out_waiting -> {:d}'.format(self.queue.qsize()))
        return self.queue.qsize()

    def _update_break_state(self):
        """\
        Set break: Controls TXD. When active, to transmitting is
        possible.
        """
        if self.logger:
            self.logger.info('_update_break_state({!r})'.format(self._break_state))

    def _update_rts_state(self):
        """Set terminal status line: Request To Send"""
        if self.logger:
            self.logger.info('_update_rts_state({!r}) -> state of CTS'.format(self._rts_state))

    def _update_dtr_state(self):
        """Set terminal status line: Data Terminal Ready"""
        if self.logger:
            self.logger.info('_update_dtr_state({!r}) -> state of DSR'.format(self._dtr_state))

    @property
    def cts(self):
        """Read terminal status line: Clear To Send"""
        if not self.is_open:
            raise PortNotOpenError()
        if self.logger:
            self.logger.info('CTS -> state of RTS ({!r})'.format(self._rts_state))
        return self._rts_state

    @property
    def dsr(self):
        """Read terminal status line: Data Set Ready"""
        if self.logger:
            self.logger.info('DSR -> state of DTR ({!r})'.format(self._dtr_state))
        return self._dtr_state

    @property
    def ri(self):
        """Read terminal status line: Ring Indicator"""
        if not self.is_open:
            raise PortNotOpenError()
        if self.logger:
            self.logger.info('returning dummy for RI')
        return False

    @property
    def cd(self):
        """Read terminal status line: Carrier Detect"""
        if not self.is_open:
            raise PortNotOpenError()
        if self.logger:
            self.logger.info('returning dummy for CD')
        return True

    # - - - platform specific - - -
    # None so far


# simple client test
if __name__ == '__main__':
    import sys
    s = Serial('loop://')
    sys.stdout.write('{}\n'.format(s))

    sys.stdout.write("write...\n")
    s.write("hello\n")
    s.flush()
    sys.stdout.write("read: {!r}\n".format(s.read(5)))

    s.close()



# -*- coding: utf-8 -*-
import json
import os

from datetime import datetime
from typing import List

import cfgtools
from old_helper import BaseRepository

from utilfunctions import (
    get_tender_info,
    is_electronic_payment,
)

from actions.models import GLAccount, TransferType, CloseOperator


class AccountRepository(BaseRepository):
    def __init__(self, mb_context):
        super(AccountRepository, self).__init__(mb_context)
        config = cfgtools.read(os.environ["LOADERCFG"])
        self.considers_tef = config.find_value("Customizations.ConsidersTEFOnBordereau", "false").lower() == "true"

    def insert_close_operator_info(self, operator):
        # type: (CloseOperator) -> float

        def inner_func(conn):
            total_bordereau_difference = 0
            initial_fund = self.get_initial_amount(operator.pos_id, operator.session)
            total_cash_in = self.get_cash_in_total_amount(operator.pos_id, operator.session)
            total_cash_out = self.get_cash_out_total_amount(operator.pos_id, operator.session)
            final_cash = initial_fund + total_cash_in - total_cash_out

            operator_values = operator.values
            self._insert_tef_tender(
                operator=operator,
                operator_values=operator_values
            )

            for value in operator_values:
                tender_id = int(value["tenderId"])
                tender = get_tender_info(int(tender_id))
                tender_is_electronic = is_electronic_payment(tender)
                amount = float(value["amount"])
                if self.considers_tef:
                    amount = float(value["amount"]) if not tender_is_electronic else None

                has_payment = str(tender_id) in operator.payments
                sold_amount = 0.0 if not has_payment else operator.payments[str(tender_id)]
                envelope = _get_envelope(tender_id, operator)
                total_bordereau_difference += _get_total_difference(
                    tender_id,
                    amount,
                    sold_amount,
                    final_cash
                )

                gl_account = GLAccount(manager=operator.manager_id, envelope=envelope)
                gl_account_json = json.dumps(gl_account, default=lambda o: o.__dict__, sort_keys=True)

                new_amount = "{}".format('NULL') if amount is None else amount

                conn.query(
                    """
                        INSERT INTO Transfer 
                        VALUES ('{}', {}, '{}', {}, '{}', {}, {}, '{}', '{}')
                    """.format(
                        operator.period,
                        operator.pos_id,
                        operator.session,
                        TransferType.DECLARED_AMOUNT,
                        "DECLARED_AMOUNT",
                        new_amount,
                        tender_id,
                        datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
                        gl_account_json
                    )
                )

                conn.query(
                    """
                        INSERT INTO Transfer 
                        VALUES ('{}', {}, '{}', {}, '{}', {}, {}, '{}', '{}')
                    """
                    .format(
                        operator.period,
                        operator.pos_id,
                        operator.session,
                        TransferType.TRANSFER_SKIM,
                        "TRANSFER_SKIM",
                        sold_amount,
                        tender_id,
                        datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
                        ""
                    )
                )

            return total_bordereau_difference

        return self.execute_with_transaction(inner_func, db_name=str(operator.pos_id))

    @staticmethod
    def _insert_tef_tender(
        operator,           # type: CloseOperator
        operator_values,    # type: List
    ):
        # type: (...) -> None

        for tender_id, amount in operator.payments.iteritems():
            tender = get_tender_info(int(tender_id))
            if not is_electronic_payment(tender):
                continue

            operator_values.append({"tenderId": int(tender_id), u"amount": str(amount)})

    def update_justification(self, pos_id, session, justification):
        def inner_func(conn):
            conn.query("""UPDATE Transfer
                          SET GLAccount = REPLACE(GLAccount, '"justification": ""', '"justification": "{}"')
                          WHERE SessionId = '{}' AND GLAccount IS NOT NULL AND GLAccount != '' AND Type IN (5,6)"""
                       .format(justification, session))

        self.execute_with_transaction(inner_func, db_name=str(pos_id))

    def get_closing_operator_sessions(self, pos_id, user, start_date, end_date):
        def inner_func(conn):
            sessions = []
            query = """SELECT SessionId
                       FROM Transfer
                       WHERE SessionId LIKE '%user={},%' and Period >= '{}' and Period <= '{}'
                       GROUP BY SessionId
                       ORDER BY Timestamp""".format(user, start_date, end_date)

            cursor = conn.select(query)
            if cursor.rows() > 0:
                for row in cursor:
                    sessions.append(row.get_entry("SessionId"))

            return sessions

        return self.execute_with_connection(inner_func, db_name=str(pos_id))

    def get_initial_amount(self, pos_id, session_id):
        return self.get_transfer_total_amount(pos_id, session_id, TransferType.INITIAL_AMOUNT)

    def get_cash_in_total_amount(self, pos_id, session_id):
        return self.get_transfer_total_amount(pos_id, session_id, TransferType.TRANSFER_CASH_IN)

    def get_cash_out_total_amount(self, pos_id, session_id):
        return self.get_transfer_total_amount(pos_id, session_id, TransferType.TRANSFER_CASH_OUT)

    def get_transfer_total_amount(self, pos_id, session_id, transfer_type):
        def inner_func(conn):
            query = """
                        SELECT amount FROM transfer
                        WHERE Type = '{}' AND SessionId = '{}'
                    """.format(transfer_type, session_id)
            cursor = conn.select(query)
            if cursor.rows() > 0:
                total_cash_in = 0
                for row in cursor:
                    total_cash_in += float(row.get_entry(0) or 0.0)

                return total_cash_in

            return 0

        return self.execute_with_connection(inner_func, db_name=str(pos_id))


def _get_envelope(tender_id, operator):
    if tender_id == 0:
        return operator.envelope
    return ""


def _get_total_difference(tender_id, amount, sold_amount, final_cash):
    if amount is None:
        amount = 0

    if tender_id == 0:
        return (amount - sold_amount) - final_cash

    return amount - sold_amount



# -*- coding: utf-8 -*-
import json

from bustoken import TK_RUPTURA_UPDATE_ITEMS
from msgbus import (
    FM_PARAM,
    TK_SYS_NAK,
)
from sysactions import (
    action,
    get_custom,
    get_model,
    get_operator_session,
    show_messagebox,
)
from typing import Any

from .. import logger
from .. import mb_context
from actions.util.get_menu_user_logged import get_menu_manager_authenticated_user
from model.customexception import UpdateRuptureException


@action
def save_rupture(
    pos_id,                 # type: str
    enabled_ingredients,    # type: str
    disabled_ingredients,   # type: str
):
    # type: (...) -> None

    model = get_model(pos_id)

    try:
        enabled_ingredients_json = json.loads(enabled_ingredients)
        disabled_codes_json = json.loads(disabled_ingredients)
        rupture_status_json = json.dumps({'enabled': enabled_ingredients_json, 'disabled': disabled_codes_json})
        last_manager_auth = get_custom(model, 'Last Manager ID') or get_menu_manager_authenticated_user(int(pos_id))
        operator = get_operator_session(model) or ''
        user_id = '{} / {}'.format(last_manager_auth, operator)
        data = '\0'.join(map(str, [rupture_status_json, user_id, pos_id]))
        msg = mb_context.MB_EasySendMessage("Ruptura", TK_RUPTURA_UPDATE_ITEMS, FM_PARAM, data)
        if msg.token == TK_SYS_NAK:
            raise UpdateRuptureException(msg.data)

        show_messagebox(pos_id, "$RUPTURE_UPDATE_SUCCESS")
    except (Exception, UpdateRuptureException) as e:
        logger.exception('[save_rupture] Exception {}'.format(e))
        show_messagebox(pos_id, "$ERROR_GETTING_DATA", buttons="$EXIT")



from application.domain import Authenticator

from sysactions import authenticate_user, AuthenticationFailed


class MsgBusAuthenticator(Authenticator):
    def authenticate(self, username, password):
        try:
            authenticate_user(username, password)
            return True
        except AuthenticationFailed:
            return False



from typing import (
    Dict,
    Optional,
)


class OrderInsertPaymentResponse(object):
    def __init__(
        self,
        tender,                     # type: Dict
    ):
        # type: (...) -> None

        self.dueAmount = tender.get("dueAmount")
        self.tenderId = tender.get("tenderId")



from production.box import ProductionBox


class OrderSpreaderBox(ProductionBox):
    def __init__(self, name, sons, logger=None):
        super(OrderSpreaderBox, self).__init__(name, sons, logger)

    def order_modified(self, order):
        for son in self.sons:
            try:
                self.debug("\n#region    - Order ID: {} - Son BOX: {} @ {}\n", order.order_id, son.name, son.tree_name)
                son.order_modified(order.clone())
            finally:
                self.debug("\n#endregion - Order ID: {} - Son BOX: {} @ {}\n", order.order_id, son.name, son.tree_name)



from logging import Logger

from mw_helper import ensure_iterable
from typing import (
    Dict,
    List,
    Union,
)

from production.model import (
    ProductionOrder,
    Item,
)
from ._OrderChangerProductionBox import OrderChangerProductionBox


class ComboTaggedWhenAllHasTagBox(OrderChangerProductionBox):
    def __init__(
        self,
        name,           # type: object
        sons,           # type: object
        tags,           # type: Union[List, str]
        tag_to_add,     # type: Union[List, str]
        logger=None,    # type: Logger
    ):
        # type: (...) -> None

        super(ComboTaggedWhenAllHasTagBox, self).__init__(name=name, sons=sons, logger=logger)
        self.tags = ensure_iterable(value=tags)
        self.tag_to_add = tag_to_add

    def change_order(
        self,
        order,  # type:  ProductionOrder
    ):
        # type: (...) ->  ProductionOrder

        if not self.tag_to_add:
            for item in order.items:
                self._add_multiple_tags(item=item)
        else:
            for item in order.items:
                if item.item_type == "COMBO" and self.all_sons_have_tags(sons=item.items):
                    item.add_tag(tag=self.tag_to_add)

        return order

    def _add_multiple_tags(
        self,
        item,   # type: Item
    ):
        # type: (...) -> None

        tags_dict = dict([(tag, True) for tag in self.tags])
        self.fill_all_sons_tags(
            sons=item.items,
            tags=tags_dict,
        )

        if item.item_type == "COMBO":
            for tag in tags_dict.keys():
                if tags_dict[tag]:
                    item.add_tag(tag=tag)

    def all_sons_have_tags(
        self,
        sons,   # type: List
    ):
        # type: (...) -> bool

        for son in sons:
            if son.item_type not in ('OPTION', 'COMBO'):
                if not son.has_some_tag(tags=self.tags):
                    return False
            else:
                all_have_tags = self.all_sons_have_tags(sons=son.items)
                if son.item_type == "COMBO" and all_have_tags:
                    son.add_tag(tag=self.tag_to_add)
                if not all_have_tags:
                    return False

        return True

    def fill_all_sons_tags(
        self,
        sons,   # type: List
        tags,   # type: Dict
    ):
        # type: (...) -> None

        for son in sons:
            if son.item_type not in ('OPTION', 'COMBO'):
                for tag in tags.keys():
                    tags[tag] = tags[tag] and son.has_tag(tag=tag)
            else:
                self.fill_all_sons_tags(
                    sons=son.items,
                    tags=tags,
                )
                if son.item_type == "COMBO":
                    for tag in tags.keys():
                        if tags[tag]:
                            son.add_tag(tag=tag)



import logging
import os

import cfgtools
from helper import import_pydevd, config_logger
from mbcontextmessagehandler import MbContextMessageBus, MbContextMessageHandler
from messagehandlerbuilder import NKDSCtrlMessageHandlerBuilder
from msgbus import MBEasyContext

def main():
    import_pydevd(os.environ["LOADERCFG"], 9135, False)
    config_logger(os.environ["LOADERCFG"], "NKDSCTRL")

    logger = logging.getLogger("NKDSCTRL")
    logger.setLevel(logging.DEBUG)
    proxy_comp_name = "KdsController"

    mb_context = MBEasyContext("NKDSCTRL")
    message_bus = MbContextMessageBus(mb_context)

    required_services = "ProductionSystem|RemoteOrder"
    message_handler = MbContextMessageHandler(
        message_bus,
        "NKDSCTRL",
        "NKDSCTRL",
        required_services,
        None
    )
    config = cfgtools.read(os.environ["LOADERCFG"])

    message_handler_builder = NKDSCtrlMessageHandlerBuilder(
        proxy_comp_name,
        mb_context,
        message_bus,
        config,
        logger
    )
    message_handler.event_handler_builder = message_handler_builder

    return message_handler.handle_events()



import json
from threading import Lock
from xml.etree import cElementTree as etree

import pyscripts
from bustoken import (
    TK_ELECTRONIC_PAYMENT_CANCEL_PAYMENT,
    TK_ELECTRONIC_PAYMENT_SEND_RESPONSE
)
from msgbus import FM_PARAM
from mw_helper import show_message_dialog, show_qr_code
from sysactions import (
    close_asynch_dialog,
    update_any_dialog,
    show_keyboard,
    show_listbox,
    show_messagebox,
)

from actions.eft import get_electronic_payment_component_name


mb_context = pyscripts.mbcontext
_current_asynchronous_dialog_id = {}
_current_asynchronous_dialog_lock = Lock()
message_lock_dict = Lock()
message_lock = {}
qrcode_dialog_id = {}


def _close_current_asynchronous_dialog(pos_id):
    dialog_id = None
    with _current_asynchronous_dialog_lock:
        if pos_id in _current_asynchronous_dialog_id:
            dialog_id = _current_asynchronous_dialog_id[pos_id]
            del _current_asynchronous_dialog_id[pos_id]

    if dialog_id is not None:
        close_asynch_dialog(pos_id, dialog_id)


def _store_current_asynchronous_dialog(pos_id, dialog_id):
    _close_current_asynchronous_dialog(pos_id)

    with _current_asynchronous_dialog_lock:
        _current_asynchronous_dialog_id[pos_id] = int(dialog_id)


def _get_current_asynchronous_dialog(pos_id):
    with _current_asynchronous_dialog_lock:
        return _current_asynchronous_dialog_id.get(pos_id, None)


def _is_electronic_payment_event(event_type):
    return event_type == "ELECTRONIC_PAYMENT"


def _send_message_response(pos_id, response):
    payment_service = get_electronic_payment_component_name(pos_id=pos_id)
    if not payment_service:
        return

    mb_context.MB_EasySendMessage(
        dest_name=payment_service,
        token=TK_ELECTRONIC_PAYMENT_SEND_RESPONSE,
        format=FM_PARAM,
        data="{}|{}".format(pos_id, response)
    )


def _show_message(pos_id, message, title):
    current_dialog_id = _get_current_asynchronous_dialog(pos_id)
    if current_dialog_id is None:
        dialog = show_message_dialog(
            pos_id,
            title,
            message,
            "NOCONFIRM",
            assync=True
        )
        _store_current_asynchronous_dialog(pos_id, dialog)
    else:
        update_any_dialog(
            pos_id, current_dialog_id, "messageDialog", message, title, "", "", "", 60000, "NOCONFIRM", "", "", "", None
        )


def _show_question(pos_id, message, title, buttons):
    _close_current_asynchronous_dialog(pos_id)
    _send_message_response(pos_id, show_messagebox(pos_id, message, title, buttons=buttons))


def _show_qrcode(pos_id, code_data, title):
    _close_current_asynchronous_dialog(pos_id)
    dialog_id = show_qr_code(pos_id, code_data, '$CANCEL', title)
    qrcode_dialog_id[pos_id] = dialog_id


def _show_input(pos_id, message, title, input_type):
    _close_current_asynchronous_dialog(pos_id)
    if not input_type or input_type == "none":
        _send_message_response(pos_id, show_messagebox(pos_id, message, title, buttons="$OK"))
    else:
        _send_message_response(pos_id, show_keyboard(pos_id, message, title, numpad=input_type == "number"))


def _show_choices(pos_id, message, title, choices):
    _close_current_asynchronous_dialog(pos_id)
    choices_dict = json.loads(choices)
    choices_keys = sorted(choices_dict.keys())
    choices_values = [choices_dict[i] for i in choices_keys]

    index = show_listbox(
        pos_id,
        choices_values,
        message=message,
        title=title,
        buttons="$OK|$CANCEL",
        icon="info",
        timeout=720000
    )
    if index is None:
        _cancel_current_payment(pos_id)

    _send_message_response(pos_id, choices_keys[index] if index is not None else "1")


def _show_numpad(pos_id, message, title):
    _close_current_asynchronous_dialog(pos_id)
    value = show_keyboard(pos_id, message, title=title, mask="INTEGER", numpad=True)
    if value is None:
        _cancel_current_payment(pos_id)
    _send_message_response(pos_id, value if value is not None else "")


def _get_pos_id_from_data(data):
    return data.split('\\0')[0]


def _get_pos_lock(data):
    with message_lock_dict:
        pos_id = _get_pos_id_from_data(data)
        if pos_id not in message_lock:
            message_lock[pos_id] = Lock()
        return message_lock[pos_id]


def _parse_event_and_call(func):
    def inner_func(params):
        data, subject, event_type = params[:3]
        if not _is_electronic_payment_event(event_type):
            return

        with _get_pos_lock(data):
            return func(*data.split('\\0'))

    return inner_func


def _dialog_resp_received(params):
    xml = params[0]
    dialog = etree.XML(xml).find("Dialog")
    dlg_id = int(dialog.get("id"))

    pos_id = None
    with _current_asynchronous_dialog_lock:
        for pos, dlg in _current_asynchronous_dialog_id.iteritems():
            if dlg == dlg_id:
                pos_id = pos
                break

    if not pos_id:
        return

    _cancel_current_payment(pos_id)
    _close_current_asynchronous_dialog(pos_id)


def _cancel_current_payment(pos_id):
    payment_service = get_electronic_payment_component_name(pos_id=pos_id)
    if not payment_service:
        return

    mb_context.MB_EasySendMessage(
        dest_name=payment_service,
        token=TK_ELECTRONIC_PAYMENT_CANCEL_PAYMENT,
        data="{}".format(pos_id)
    )


def _close_dialogs(pos_id):
    _close_current_asynchronous_dialog(pos_id)


def _close_qrcode(pos_id):
    dialog_id = qrcode_dialog_id[pos_id]
    close_asynch_dialog(pos_id, dialog_id)
    del qrcode_dialog_id[pos_id]


def init():
    pyscripts.subscribe_event_listener("SHOW_MESSAGE", _parse_event_and_call(_show_message))
    pyscripts.subscribe_event_listener("SHOW_QUESTION", _parse_event_and_call(_show_question))
    pyscripts.subscribe_event_listener("SHOW_QRCODE", _parse_event_and_call(_show_qrcode))
    pyscripts.subscribe_event_listener("SHOW_INPUT", _parse_event_and_call(_show_input))
    pyscripts.subscribe_event_listener("SHOW_CHOICES", _parse_event_and_call(_show_choices))
    pyscripts.subscribe_event_listener("SHOW_NUMPAD", _parse_event_and_call(_show_numpad))
    pyscripts.subscribe_event_listener("CLOSE_DIALOGS", _parse_event_and_call(_close_dialogs))
    pyscripts.subscribe_event_listener("CLOSE_QRCODE", _parse_event_and_call(_close_qrcode))
    pyscripts.subscribe_event_listener("DIALOG_RESP", _dialog_resp_received)



from logging import Logger

from interactor.route.menu.get_images.dto import (
    GetImagesRequest,
    GetImagesResponse
)
from model.config import Config
from model.services import ImagesServiceModel
from validation_util import validate_api_key


class GetImagesInteractor(object):

    def __init__(self, logger, config, images_service):
        # type: (Logger, Config, ImagesServiceModel) -> None

        self.logger = logger
        self.config = config
        self.images_service = images_service

    def execute(self, request):
        # type: (GetImagesRequest) -> GetImagesResponse

        validate_api_key(request, self.config.api_key_mapping)

        return self.images_service.get_images(
            images_path=self.config.images_path,
            totem_config_path=self.config.totem_config_path
        )



from commons.dto import ReportColumnDefinition, AlignTypes
from commons.report import Formatter
from commons.util import TableReport, DefaultHeaderFormatter
from reports_app.cashtransferreport.dto import CashTransferReportDto  # noqa


class CashTransferFormatter(Formatter):
    def __init__(self):
        super(CashTransferFormatter, self).__init__()
        self.translated_messages = {}

    def format_report(self, cash_transfer_report_dto, translator):
        # type: (CashTransferReportDto) -> None
        self.translated_messages = translator.translate_labels(["$CASH_OUT_TITLE",
                                                                "$CASH_IN_TITLE",
                                                                "$BUSINESS_DAY",
                                                                "$STORE",
                                                                "$POS",
                                                                "$OPERATOR",
                                                                "$PERIOD",
                                                                "$CURRENT_DATE",
                                                                "$TYPE",
                                                                "$OP",
                                                                "$HOUR",
                                                                "$VALUE",
                                                                "$MANAGER_MENU",
                                                                "$LOGOUT",
                                                                "$LOGIN",
                                                                "$MANAGER_MENU_CASH_TRANSFER_REPORT"
                                                                ],
                                                               cash_transfer_report_dto.pos_id)

        report = self._append_header(cash_transfer_report_dto, translator)
        report += self._append_body(cash_transfer_report_dto)
        report_bytes = report.decode("utf-8").encode("utf-8")
        return report_bytes

    def _append_body(self, cash_transfer_report_dto):
        report_body_lines = []

        self._append_transfers_title(report_body_lines)
        self._append_transfers(cash_transfer_report_dto, report_body_lines)
        self._append_line_separator(report_body_lines)
        return self._format_body(report_body_lines)

    @staticmethod
    def _append_line_separator(report_body_lines):
        report_body_lines.append([ReportColumnDefinition(
            fill_with_char='='
        )])

    @staticmethod
    def _format_body(report_body_lines):
        table_formatter = TableReport([
            ReportColumnDefinition(text='', width=15, fill_with_char='', before_text='', after_text='',
                                   align=AlignTypes.LEFT),
            ReportColumnDefinition(text='', width=7, fill_with_char='', before_text='', after_text='',
                                   align=AlignTypes.LEFT),
            ReportColumnDefinition(text='', width=6, fill_with_char='', before_text='', after_text='',
                                   align=AlignTypes.LEFT),
            ReportColumnDefinition(text='', width=10, fill_with_char='', before_text='', after_text='',
                                   align=AlignTypes.RIGHT)
        ])
        return table_formatter.format(report_body_lines)

    def _append_transfers(self, cash_transfer_report_dto, report_body_lines):
        for all_transfers in cash_transfer_report_dto.all_transfers_dto:
            self._append_transfers_date(all_transfers, report_body_lines)

            for transfer in all_transfers.transfers_by_date:
                self._append_transfers_info(transfer, report_body_lines)

    def _append_transfers_info(self, transfer, report_body_lines):
        report_body_lines.append(
            [self.translated_messages[transfer.transfer_type].upper(),
             transfer.transfer_operator,
             transfer.transfer_date.strftime("%H:%M"),
             (str(transfer.transfer_value))])

    @staticmethod
    def _append_transfers_date(transfers_by_day, report_body_lines):
        report_body_lines.append([ReportColumnDefinition(
            text=transfers_by_day.day,
            fill_with_char='-',
            before_text=' ',
            after_text=' ',
            align=AlignTypes.CENTER
        )])

    def _append_header(self, cash_transfer_report_dto, translator):
        default_header_formatter = DefaultHeaderFormatter(translator)
        if cash_transfer_report_dto.header_transfer_dto.report_type == 'cash_out':
            title = "{} ({})".format(self.translated_messages["$CASH_OUT_TITLE"],
                                     self.translated_messages["$BUSINESS_DAY"].upper())
        else:
            title = "{} ({})".format(self.translated_messages["$CASH_IN_TITLE"],
                                     self.translated_messages["$BUSINESS_DAY"].upper())

        return default_header_formatter.format_header(
            title, cash_transfer_report_dto.header_transfer_dto, cash_transfer_report_dto.pos_id)

    def _append_transfers_title(self, report_body_lines):
        report_body_lines.append([self.translated_messages["$TYPE"],
                                  self.translated_messages["$OP"],
                                  self.translated_messages["$HOUR"],
                                  self.translated_messages["$VALUE"] + "(R$)"])



import os
import json
import logging.config

from abc import ABCMeta

from helper import config_logger
from typing import Callable, Any, List

from msgbus import MBEasyContext
from persistence import Connection, Driver
from systools import sys_parsefilepath


class BaseRepository(object):
    __metaclass__ = ABCMeta

    def __init__(self, mb_context):
        # type: (MBEasyContext) -> None
        self.mb_context = mb_context

    def execute_with_connection(self, method, db_name=None, service_name=None):
        # type: (Callable[[Connection], Any], int, str) -> Any
        conn = None
        try:
            conn = Driver().open(self.mb_context,
                                 dbname=str(db_name) if db_name is not None else None,
                                 service_name=service_name)
            return method(conn)
        finally:
            if conn is not None:
                conn.close()

    def execute_with_transaction(self, method, db_name=None, service_name=None):
        # type: (Callable[[Connection], Any], int, str) -> Any
        def transaction_method(conn):
            # type: (Connection) -> Any

            conn.transaction_start()
            try:
                conn.query("BEGIN TRANSACTION")
                ret = method(conn)
                conn.query("COMMIT")
                return ret
            except:
                conn.query("ROLLBACK")
                raise
            finally:
                conn.transaction_end()

        return self.execute_with_connection(transaction_method, db_name=db_name, service_name=service_name)

    def execute_in_all_databases(self, method, pos_list):
        # type: (Callable[[Connection], Any], List[int]) -> Any
        ret = []
        for pos in pos_list:
            conn = None
            try:
                conn = Driver().open(self.mb_context, dbname=str(pos))
                ret.append(method(conn))
            finally:
                if conn is not None:
                    conn.close()

        return ret

    def execute_in_all_databases_returning_flat_list(self, method, pos_list):
        # type: (Callable[[Connection], Any], List[int]) -> Any
        inner_ret = self.execute_in_all_databases(method, pos_list)

        ret = []
        for inner_list in inner_ret:
            ret.extend(inner_list)

        return ret


def find_value_with_vars(cfg, path, default_value=None):
    value = cfg.find_value(path, default_value)
    if value is None:
        return value

    for variable_name in os.environ:
        variable_format = "{{{}}}".format(variable_name)
        if variable_format in value:
            value = value.replace(variable_format, os.environ[variable_name])
            if variable_name in ("BUNDLEDIR", "HVDATADIR"):
                value = value.replace("/./", os.sep).replace("//", os.sep).replace("/", os.sep)

    return value


def config_logger_parse_file_name(loader_path,
                                  log_name='component',
                                  configure_root=False,
                                  root_level=logging.WARN,
                                  max_files=5,
                                  max_size=5242880,
                                  add_error_handler=False):

    real_path = os.path.realpath(loader_path)
    if os.path.isfile(real_path) and real_path.endswith(".json"):
        with open(real_path, "rb") as f:
            json_loads = json.loads(f.read().decode("utf-8"))
            handlers = json_loads.get("handlers")
            if handlers:
                for handler in handlers:
                    file_name = sys_parsefilepath(handlers[handler].get("filename", ""))
                    if file_name:
                        handlers[handler]["filename"] = file_name

                loader_path = json_loads

    config_logger(loader_path, log_name, configure_root, root_level, max_files, max_size, add_error_handler)





from abc import ABCMeta, abstractmethod

from domain import Session


class SessionRetriever(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_session(self, id):
        # type: (str) -> Session
        raise NotImplementedError()



import json
from typing import (
    Any,
    Optional,
)

from application.interactor import GetAndLockVoucherInteractor
from application.model import AsyncUuidMessageProcessor
from application.model.configuration import Configurations
from messagebus import (
    DefaultToken,
    DataType,
    Message,
    MessageBus,
)


class GetAndLockVoucherProcessor(AsyncUuidMessageProcessor):
    def __init__(self, configs, interactor):
        # type: (Configurations, GetAndLockVoucherInteractor) -> None

        super(GetAndLockVoucherProcessor, self).__init__()
        self.logger = configs.logger
        self.configs = configs
        self.interactor = interactor

    def parse_data(self, data):
        # type: (str) -> Any

        parsed_data = json.loads(data)
        pos_id = int(parsed_data.get("posId"))
        voucher_id = str(parsed_data.get("voucherId"))
        return pos_id, voucher_id

    def call_business(self, params):
        # type: ((int, str)) -> str

        pos_id, voucher_id = params

        return self.interactor.get_and_lock_voucher(pos_id, voucher_id)

    def format_response(self, message_bus, message, event, request, response):
        if response:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_ACK, DataType.string, response))
        else:
            message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK))

    def format_exception(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Any]
        input_obj,      # type: Optional[Any]
        exception,      # type: Exception
    ):
        # type: (...) -> None

        data = exception.message if exception.message else str(exception)
        message_bus.reply_message(
            message=message,
            reply_message=Message(
                token=DefaultToken.TK_SYS_NAK,
                data_type=DataType.param,
                data=data,
            ),
        )

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))



from ..utils import cached_property
from .auth import WWWAuthenticateMixin
from .base_response import BaseResponse
from .common_descriptors import CommonResponseDescriptorsMixin
from .cors import CORSResponseMixin
from .etag import ETagResponseMixin


class ResponseStream(object):
    """A file descriptor like object used by the :class:`ResponseStreamMixin` to
    represent the body of the stream.  It directly pushes into the response
    iterable of the response object.
    """

    mode = "wb+"

    def __init__(self, response):
        self.response = response
        self.closed = False

    def write(self, value):
        if self.closed:
            raise ValueError("I/O operation on closed file")
        self.response._ensure_sequence(mutable=True)
        self.response.response.append(value)
        self.response.headers.pop("Content-Length", None)
        return len(value)

    def writelines(self, seq):
        for item in seq:
            self.write(item)

    def close(self):
        self.closed = True

    def flush(self):
        if self.closed:
            raise ValueError("I/O operation on closed file")

    def isatty(self):
        if self.closed:
            raise ValueError("I/O operation on closed file")
        return False

    def tell(self):
        self.response._ensure_sequence()
        return sum(map(len, self.response.response))

    @property
    def encoding(self):
        return self.response.charset


class ResponseStreamMixin(object):
    """Mixin for :class:`BaseResponse` subclasses.  Classes that inherit from
    this mixin will automatically get a :attr:`stream` property that provides
    a write-only interface to the response iterable.
    """

    @cached_property
    def stream(self):
        """The response iterable as write-only stream."""
        return ResponseStream(self)


class Response(
    BaseResponse,
    ETagResponseMixin,
    WWWAuthenticateMixin,
    CORSResponseMixin,
    ResponseStreamMixin,
    CommonResponseDescriptorsMixin,
):
    """Full featured response object implementing the following mixins:

    -   :class:`ETagResponseMixin` for etag and cache control handling
    -   :class:`WWWAuthenticateMixin` for HTTP authentication support
    -   :class:`~werkzeug.wrappers.cors.CORSResponseMixin` for Cross
        Origin Resource Sharing headers
    -   :class:`ResponseStreamMixin` to add support for the ``stream``
        property
    -   :class:`CommonResponseDescriptorsMixin` for various HTTP
        descriptors
    """



from application.model.configuration import Configurations


class TestConnectionInteractor(object):

    def __init__(self, configs):
        # type: (Configurations) -> None

        self.configs = configs
        self.logger = self.configs.logger

    def test_connection(self, current_pos_id):
        # type: (int) -> bool

        self.logger.info("Testing nCoupon connection. CurrentPosId: {}".format(current_pos_id))
        return True



from datetime import datetime
from sqlite3 import Connection

from domain import TipRepository, OperatorTip
from old_helper import BaseRepository
from typing import List


class MwTipRepository(TipRepository, BaseRepository):

    def get_tips_per_operator(self, start_date=None, end_date=None, operator=None):
        # type: (datetime, datetime, str) -> List[OperatorTip]
        def inner_func(conn):
            # type: (Connection) -> List[OperatorTip]
            query = self._TipsQuery.format(
                start_date.strftime("%Y-%m-%d %H:%M:%S"),
                end_date.strftime("%Y-%m-%d %H:%M:%S"),
                " AND oicp.Value = {0}".format(operator) if operator is not None else ""
            )

            return [OperatorTip(x.get_entry(0), int(x.get_entry(1)), float(x.get_entry(2)), float(x.get_entry(3))) for x in conn.select(query)]

        return self.execute_with_connection(inner_func)

    _TipsQuery = """
        SELECT oicp.Value,
               sp.ServiceId,
               p.DefaultUnitPrice,
               tip.tip
        FROM ServiceSplit sp
        INNER JOIN OrderItem oi ON sp.SplitOrderId = oi.OrderId
        AND sp.SplitLineNumber = oi.LineNumber
        AND oi.OrderedQty > 0
        INNER JOIN
          (SELECT sp.OrderId,
                  sp.LineNumber,
                  Value
           FROM ServiceSplit sp
           INNER JOIN OrderItemCustomProperties oicp ON sp.OrderId = oicp.OrderId
           AND sp.LineNumber = oicp.LineNumber
           AND oicp.Key = 'operator' {2}
        ) oicp ON sp.OrderId = oicp.OrderId
        AND sp.LineNumber = oicp.LineNumber
        INNER JOIN Price p ON oi.PriceKey = p.PriceKey
        INNER JOIN
          (SELECT o.OrderId,
                  max(Tip) / max(TotalGross) tip
           FROM TableService ts
           INNER JOIN ServiceSplit sp ON ts.serviceId = sp.ServiceId
           INNER JOIN Orders o ON sp.SplitOrderId = o.OrderId
           INNER JOIN ServiceTip st ON sp.ServiceId = st.ServiceId
           AND sp.SplitKey = st.SplitKey
           WHERE ts.StartTS >= '{0}' and ts.FinishedTS <= '{1}'
        GROUP BY o.OrderId) tip ON oi.OrderId = tip.OrderId
    """



import time
import uuid
from logging import Logger
from xml.etree import cElementTree as eTree
from threading import Thread

import cfgtools
from typing import List
from xml.etree.ElementTree import Element

from messagehandler import MessageHandler
from messagebus import TokenCreator, TokenPriority, AckMessage, DefaultToken, Event, Message, NakMessage, MessageBus, \
    MessageBusException
from msgbus import (
    FM_PARAM,
    MBEasyContext
)
from model import (
    Kds,
    KdsModel,
    KdsView
)

from services import DeliveryService

MSG_GRP_KDS = '0B'
TK_KDS_GETMODEL = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "1")
TK_KDS_UPDATE_VIEW = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "2")
TK_KDS_REFRESH = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "3")
TK_KDS_REFRESH_END = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "4")
TK_KDS_SET_PROD_STATE = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "5")
TK_KDS_UNDO_SERVE = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "6")
TK_KDS_TOGGLE_TAG_LINE = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "8")
TK_KDS_CHANGE_VIEW = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "A")
TK_KDS_BUMP_BAR_COMMAND = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "B")
TK_KDS_BUMP_ZOOM_NEXT = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "C")
TK_KDS_RELOAD = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "14")
TK_KDS_SELECT_ORDER = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "20")
TK_KDS_UPDATE_STATISTICS = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "21")

MSG_GRP_PRODUCTION = "0C"
TK_PROD_REFRESH_VIEW = TokenCreator.create_token(TokenPriority.low, MSG_GRP_PRODUCTION, "1")
TK_PROD_SET_ORDER_STATE = TokenCreator.create_token(TokenPriority.low, MSG_GRP_PRODUCTION, "2")
TK_PROD_UNDO = TokenCreator.create_token(TokenPriority.low, MSG_GRP_PRODUCTION, "7")
TK_PROD_TOGGLE_TAG_LINE = TokenCreator.create_token(TokenPriority.low, MSG_GRP_PRODUCTION, "8")
TK_PROD_GET_STATISTICS = TokenCreator.create_token(TokenPriority.low, MSG_GRP_PRODUCTION, "102")

new_production = "ProductionSystem"

none_kds_found_message = "None KDS found"


class ProxyMessageHandler(MessageHandler):
    def __init__(
        self,
        comp_name,    # type: str
        mb_context,   # type: MBEasyContext
        message_bus,  # type: MessageBus
        config,       # type: cfgtools.Group
        logger        # type: Logger
    ):
        # type: (...) -> None

        self.comp_name = comp_name
        self.kds_list = None
        self.mb_context = mb_context
        self.message_bus = message_bus
        self.logger = logger
        self.commands_uuid = []
        self.config = config

        self.statistics_cache = ''
        init_thread = Thread(target=self.get_current_statistics)
        init_thread.daemon = True
        init_thread.start()

    def handle_event(self, message_bus, event):
        return

    def handle_message(self, message_bus, message):
        data = message.data.split('\0')
        resp = None

        if message.token == TK_KDS_SELECT_ORDER:
            kds_id = int(data[0])
            self.logger.info("Receiving TK_KDS_SELECT_ORDER for KDS - {:02d}".format(kds_id))

            # TODO: Change Selected Order and Store it

            message.token = DefaultToken.TK_SYS_ACK

        elif message.token == TK_KDS_GETMODEL:
            resp = self.handle_kds_get_model(data)
        elif message.token == TK_KDS_CHANGE_VIEW:
            resp = self.handle_kds_change_view(data)
        elif message.token == TK_KDS_UPDATE_VIEW:
            resp = self.handle_kds_update_view(data, message_bus)
        elif message.token == TK_KDS_REFRESH:
            resp = self.handle_kds_refresh(data, message_bus)
        elif message.token == TK_KDS_REFRESH_END:
            resp = self.handle_kds_refresh_end(data, message_bus)
        elif message.token == TK_KDS_SET_PROD_STATE:
            resp = self.handle_kds_set_prod_state(data, message_bus)
        elif message.token == TK_KDS_UNDO_SERVE:
            resp = self.handle_kds_undo_serve(data, message_bus)
        elif message.token == TK_KDS_UPDATE_STATISTICS:
            resp = self.handle_kds_update_statistics(data, message_bus)
        elif message.token == TK_KDS_BUMP_BAR_COMMAND:
            resp = self.handle_kds_bump_bar_command(data, message_bus)
        elif message.token == TK_KDS_BUMP_ZOOM_NEXT:
            resp = self.handle_kds_zoom_next(data)
        elif message.token == TK_KDS_TOGGLE_TAG_LINE:
            resp = self.handle_kds_toggle_tag_line(data, message_bus)
        elif message.token == TK_KDS_RELOAD:
            resp = self.handle_kds_reload(data, message_bus)
        else:
            resp = self.forward_message(message, message_bus)

        message_bus.reply_message(message, resp)
        self.logger.info("Message replied")

    def handle_kds_get_model(self, data):
        self.logger.info("TK_KDS_GETMODEL - Received")

        kds_id = data[0]
        kds = self.get_kds(kds_id)

        if not kds:
            self.logger.info("TK_KDS_GETMODEL - KdsId: {} not found".format(kds_id))
            return NakMessage(none_kds_found_message, 0)

        self.logger.info("TK_KDS_GETMODEL - KdsId: {}".format(kds_id))

        kds_model = KdsModel(kds, kds_id)
        kds_model_xml = kds_model.get_model_xml()
        kds_model_xml.append(eTree.Element("Statistics", text=self.statistics_cache))
        resp = AckMessage(eTree.tostring(kds_model_xml, encoding='utf8', method='xml'), 0)

        self.logger.info("TK_KDS_GETMODEL - Finished")
        return resp

    def handle_kds_change_view(self, data):
        self.logger.info("TK_KDS_CHANGE_VIEW - Received")

        kds_id = data[0]
        view_name = data[1]
        cmd_uuid = data[2] if len(data) > 2 else '0'
        kds = self.get_kds(kds_id)

        if not kds:
            self.logger.info("TK_KDS_CHANGE_VIEW - KdsId: {} not found".format(kds_id))
            return NakMessage(none_kds_found_message, 0)

        if self._is_uuid_unique(cmd_uuid):
            self.set_selected_view(kds, view_name)
            self.logger.info("TK_KDS_CHANGE_VIEW - KdsId: {}".format(kds_id))

        kds_model = KdsModel(kds, kds_id)
        kds_model_xml = kds_model.get_model_xml()
        resp = AckMessage(eTree.tostring(kds_model_xml, encoding='utf8', method='xml'), 0)

        self.logger.info("TK_KDS_CHANGE_VIEW - Finished")
        return resp

    def get_kds(self, kds_id):
        return next((x for x in self._get_kds_list() if x.name == kds_id), None)

    @staticmethod
    def get_bump_bar_command(bump_bar_commands, command_code):
        return next((x for x in bump_bar_commands if command_code in x.keys), None)

    @staticmethod
    def set_selected_view(kds, view_name):
        for view in kds.views:
            view.selected = view.name == view_name

        has_selected_view = next((x for x in kds.views if x.selected), None)
        if not has_selected_view:
            kds.views[0].selected = True

    @staticmethod
    def set_view_zoom(kds, production_view):
        view = next((x for x in kds.views if x.production_view == production_view), None)
        if not view:
            return

        view.selected_zoom += 1
        if view.selected_zoom == len(view.zoom_levels):
            view.selected_zoom = 0

    def handle_kds_update_view(self, data, message_bus):
        self.logger.info("TK_KDS_UPDATE_VIEW - Received")

        production_view = data[0]
        xml = eTree.XML(data[1])
        order_id = xml.get("order_id")

        self.logger.info("TK_KDS_UPDATE_VIEW - ProductionView: {} / OrderId: {}".format(production_view, order_id))

        view_update_xml = self.create_view_update_xml(production_view, xml)
        for kds in self.get_kds_list_with_production_view(production_view):
            self.logger.info("TK_KDS_UPDATE_VIEW - Sending event for KDS: {} / OrderId: {}".format(kds, order_id))

            event_subject = "KDS" + kds
            event_type = "PROD_ORDER_" + order_id + "_" + production_view
            event = self.send_event(event_subject, event_type, view_update_xml)
            message_bus.publish_event(event)

        self.logger.info("TK_KDS_UPDATE_VIEW - Finished")

        return AckMessage('', 0)

    def handle_kds_refresh(self, data, message_bus):
        self.logger.info("TK_KDS_REFRESH - Received")
        response = AckMessage('', 0)

        kds_id = data[0]
        for view in self.get_production_views_with_kds_id(kds_id):
            self.logger.info("TK_KDS_REFRESH - Sending event for View: {}".format(view))

            message = Message(TK_PROD_REFRESH_VIEW, FM_PARAM, str(view), 3000000)
            reply = message_bus.send_message(new_production, message)
            if reply.token != DefaultToken.TK_SYS_ACK:
                response = NakMessage()

        self.logger.info("TK_KDS_REFRESH - Finished")
        return response

    def handle_kds_set_prod_state(self, data, message_bus):
        self.logger.info("TK_KDS_SET_PROD_STATE - Received")

        kds_id = data[0]
        order_id = data[1]
        prod_state = data[2]
        view = data[3]
        cmd_uuid = data[4] if len(data) > 4 else '0'

        if self._is_uuid_unique(cmd_uuid):
            log_message = "TK_KDS_SET_PROD_STATE - Sending event to KDS: {} / OrderId: {} / ProdState {}"
            self.logger.info(log_message.format(kds_id, order_id, prod_state))

            message_data = order_id + "\0" + view + "\0" + prod_state
            message = Message(TK_PROD_SET_ORDER_STATE, FM_PARAM, message_data, 3000000)
            message_bus.send_message(new_production, message)

        self.logger.info("TK_KDS_SET_PROD_STATE - Finished")
        return AckMessage('', 0)

    def handle_kds_zoom_next(self, data):
        self.logger.info("TK_KDS_BUMP_ZOOM_NEXT - Received")

        kds_id = data[0]
        production_view = data[1]
        cmd_uuid = data[2] if len(data) > 2 else '0'
        kds = self.get_kds(kds_id)

        if not kds:
            self.logger.info("TK_KDS_BUMP_ZOOM_NEXT - KdsId: {} not found".format(kds_id))
            return NakMessage(none_kds_found_message, 0)

        if self._is_uuid_unique(cmd_uuid):
            self.set_view_zoom(kds, production_view)
            self.logger.info("TK_KDS_BUMP_ZOOM_NEXT - KdsId: {}".format(kds_id))

        kds_model = KdsModel(kds, kds_id)
        kds_model_xml = kds_model.get_model_xml()
        resp = AckMessage(eTree.tostring(kds_model_xml, encoding='utf8', method='xml'), 0)

        self.logger.info("TK_KDS_BUMP_ZOOM_NEXT - Finished")
        return resp

    def handle_kds_refresh_end(self, data, message_bus):
        self.logger.info("TK_KDS_REFRESH_END - Received")

        production_view = data[0]
        for kds in self.get_kds_list_with_production_view(production_view):
            self.logger.info("TK_KDS_REFRESH_END - Sending event to KDS: {} - VIEW: {}".format(kds, production_view))

            event_subject = "KDS" + kds
            event_type = "REFRESH_END_" + production_view
            refresh_end_xml = eTree.XML("<RefreshEnd name=\"{}\" updated=\"true\"/>".format(production_view))
            event = self.send_event(event_subject, event_type, refresh_end_xml)
            message_bus.publish_event(event)

        self.logger.info("TK_KDS_REFRESH_END - Finished")

        return AckMessage('', 0)

    def handle_kds_reload(self, data, message_bus):
        self.logger.info("TK_KDS_RELOAD - Received")

        production_view = data[0]
        for kds in self.get_kds_list_with_production_view(production_view):
            self.logger.info("TK_KDS_RELOAD - Sending event to KDS: {}".format(kds))

            event_subject = "KDS" + kds
            event_type = "RELOAD_" + production_view
            refresh_end_xml = eTree.XML("<Reload name=\"{}\"/>".format(production_view))
            event = self.send_event(event_subject, event_type, refresh_end_xml)
            message_bus.publish_event(event)

        self.logger.info("TK_KDS_RELOAD - Finished")

        return AckMessage('', 0)

    def handle_kds_toggle_tag_line(self, data, message_bus):
        self.logger.info("TK_KDS_TOGGLE_TAG_LINE - Received")

        kds_id = data[0]
        order_id = data[1]
        line_number = data[2]
        tag_name = data[3]
        view = data[4]
        cmd_uuid = data[5] if len(data) > 5 else '0'

        if self._is_uuid_unique(cmd_uuid):
            log_message = "TK_KDS_TOGGLE_TAG_LINE - Sending event to KDS: {} / OrderId: {} / LineNumber {} / TagName {}"
            self.logger.info(log_message.format(kds_id, order_id, line_number, tag_name))

            message_data = order_id + "\0" + view + "\0" + line_number + "\0" + tag_name
            message = Message(TK_PROD_TOGGLE_TAG_LINE, FM_PARAM, message_data, 3000000)
            message_bus.send_message(new_production, message)

        self.logger.info("TK_KDS_TOGGLE_TAG_LINE - Finished")
        return AckMessage('', 0)

    def handle_kds_update_statistics(self, data, message_bus):
        self.logger.info("TK_KDS_UPDATE_STATISTICS - Received")

        try:
            statistics = data[0]
            self.statistics_cache = statistics
        except ValueError:
            pass

        for kds_id in self.get_all_kds_ids():
            event_data = eTree.tostring(self.create_statistics_event_from_json(self.statistics_cache))
            event = Event(subject='KDS{}'.format(kds_id), event_type='Statistics', data=event_data)
            message_bus.publish_event(event)

        self.logger.info("TK_KDS_UPDATE_STATISTICS - Finished")

        return AckMessage('', 0)

    def handle_kds_bump_bar_command(self, data, message_bus):
        self.logger.info("TK_KDS_BUMP_BAR_COMMAND - Received")

        kds_id = data[0]
        bump_bar_name = data[1]
        command_code = data[2]

        log_message = "TK_KDS_BUMP_BAR_COMMAND - KdsId: {} / BumpBar: {} / Command: {}"
        self.logger.info(log_message.format(kds_id, bump_bar_name, command_code))

        kds = self.get_kds(kds_id)
        if not kds:
            self.logger.info("TK_KDS_BUMP_BAR_COMMAND - KdsId: {} not found".format(kds_id))
            return NakMessage(none_kds_found_message, 0)

        bump_bar = kds.bump_bar
        command = self.get_bump_bar_command(bump_bar.commands, command_code)
        if not command:
            self.logger.info("TK_KDS_BUMP_BAR_COMMAND - Key not mapped")
            return AckMessage('', 0)

        root_element = eTree.Element("Event")
        eTree.SubElement(root_element, "Command", name=command.name, value=str(uuid.uuid4()))
        event = Event(subject='KDS{}'.format(kds_id), event_type='KdsModel', data=eTree.tostring(root_element))
        message_bus.publish_event(event)

        self.logger.info("TK_KDS_BUMP_BAR_COMMAND - Received")
        return AckMessage('', 0)

    def handle_kds_undo_serve(self, data, message_bus):
        self.logger.info("TK_KDS_UNDO_SERVE - Received")

        kds_id = data[0]
        view = data[1]
        cmd_uuid = data[2] if len(data) > 2 else '0'

        if self._is_uuid_unique(cmd_uuid):
            log_message = "TK_KDS_UNDO_SERVE - Sending event to KDS: {}"
            self.logger.info(log_message.format(kds_id))

            message = Message(TK_PROD_UNDO, FM_PARAM, str(view), 3000000)
            message_bus.send_message(new_production, message)

        self.logger.info("TK_KDS_UNDO_SERVE - Finished")
        return AckMessage('', 0)

    @staticmethod
    def create_view_update_xml(production_view, xml):
        view_update_xml = eTree.Element("ViewUpdate", name=production_view)
        view_update_xml.append(xml)

        return view_update_xml

    @staticmethod
    def send_event(event_subject, event_type, xml):
        event_xml = eTree.Element("Event", subject=event_subject, type=event_type)
        event_xml.append(xml)
        event = Event(event_subject, event_type, eTree.tostring(event_xml))

        return event

    def forward_message(self, message, message_bus):
        resp = NakMessage('', 0)
        try:
            self.logger.info("Receiving message for KDS - data = {}".format(message))
            resp = message_bus.send_message(self.comp_name, message)
            self.logger.info("Message forwarded")
        except MessageBusException: # noqa
            self.logger.info("Error forwarding message - Component not found")

        return resp

    def get_kds_list_with_production_view(self, kds_view):
        kds_list = []
        for kds in self._get_kds_list():
            if len([v for v in kds.views if v.production_view == kds_view]) > 0:
                kds_list.append(kds.name)

        return list(set(kds_list))

    def get_production_views_with_kds_id(self, kds_id):
        view_list = []

        kds = next((x for x in self._get_kds_list() if x.name == kds_id), None)
        for view in kds.views:
            view_list.append(view.production_view)

        return list(set(view_list))

    @staticmethod
    def create_statistics_event_from_json(json_data):
        # type: (str) -> Element
        root_element = eTree.Element("Event")
        eTree.SubElement(root_element, "Statistics", text=json_data)
        return root_element

    def get_all_kds_ids(self):
        kds_list = []
        for kds in self._get_kds_list():
            kds_list.append(kds.name)

        return list(set(kds_list))

    def get_current_statistics(self):
        statistics_ok = False
        while statistics_ok is False:
            try:
                resp = self.message_bus.send_message(new_production, Message(TK_PROD_GET_STATISTICS))
                if resp.token == DefaultToken.TK_SYS_ACK:
                    self.statistics_cache = resp.data
                    statistics_ok = True
            except BaseException: # noqa
                time.sleep(0.500)

    def _is_uuid_unique(self, cmd_uuid):
        # type: (str) -> bool

        self.logger.info('UUID: {}'.format(cmd_uuid))
        if cmd_uuid != '0' and cmd_uuid in self.commands_uuid:
            self.logger.info('Command Ignored')
            return False

        self.commands_uuid.append(cmd_uuid)
        self.commands_uuid = self.commands_uuid[-5:]
        return True

    def _get_kds_list(self):
        # type: (...) -> List[Kds]

        if self.kds_list is not None:
            return self.kds_list

        kds_views = self._get_kds_views()
        kds_list = []
        kds_config = self.config.find_group("KDS")
        if not kds_config:
            return []

        delivery_service = DeliveryService(self.mb_context, self.logger)
        delivery_brands = delivery_service.get_delivery_brands()
        for kds_config_group in kds_config.groups:
            kds_list.append(
                Kds(
                    kds_config_group,
                    kds_views,
                    delivery_brands
                )
            )

        self.kds_list = kds_list
        return self.kds_list

    def _get_kds_views(self):
        # type: (...) -> List[KdsView]

        views = []
        view_config = self.config.find_group("Views")
        if not view_config:
            return []

        for view_config_group in view_config.groups:
            views.append(
                KdsView(view_config_group)
            )

        return views



from __future__ import absolute_import
from future.utils import PY3

if PY3:
    from dbm import *
else:
    __future_module__ = True
    from whichdb import *
    from anydbm import *

# Py3.3's dbm/__init__.py imports ndbm but doesn't expose it via __all__.
# In case some (badly written) code depends on dbm.ndbm after import dbm,
# we simulate this:
if PY3:
    from dbm import ndbm
else:
    try:
        from future.moves.dbm import ndbm
    except ImportError:
        ndbm = None



from helper import remove_accents
from systools import sys_log_exception
from typing import (
    List,
)

from basereport import (
    SimpleReport,
    TableReport,
    ReportColumnDefinition,
    AlignTypes,
)
from report import (
    Presenter,
    Part,
)
from report.command import (
    NewLineCommand,
    RepeatCommand,
    FontCommand,
    AlignCommand,
)

from reportsdata.justificationreport.dto import (
    JustificationGeneratorDto,
    JustificationBodyDto,
    JustificationHeaderDto,
)
from utils import format_line


class JustificationReportPresenter(Presenter):

    def __init__(
        self,
        columns,    # type: int
    ):
        # type: (...) -> None

        self.total = 0
        self.columns = columns

        self.header_column_definition = [
            ReportColumnDefinition(
                width=columns,
                align=AlignTypes.CENTER,
            )
        ]

        self.body_definitions = [
            ReportColumnDefinition(width=16, after_fill_text=': ', fill_with_char=''),
            ReportColumnDefinition(width=columns - 16)
        ]

        self.center_definitions = [
            ReportColumnDefinition(width=columns),
        ]

        self.footer_column_definition = [
            ReportColumnDefinition(
                width=columns,
                align=AlignTypes.CENTER,
            )
        ]

    def present(
        self,
        dto,     # type: JustificationGeneratorDto
    ):
        # type: (...) -> SimpleReport

        try:
            parts = [Part(commands=[FontCommand(FontCommand.Font.B)])]
            parts.extend(self._build_header(dto.header))
            parts.extend(self._build_body(dto.body))
            parts.extend(self._build_footer())
            parts.extend([Part(None, [NewLineCommand()])])
        except Exception as _:
            sys_log_exception("Error building Benefits Report")
            raise

        return SimpleReport(
            parts=parts,
            width=self.columns,
        )

    def _build_header(
        self,
        header,      # type: JustificationHeaderDto
    ):
        # type: (...) -> List

        parts = [
            self._get_title(header.title)[0],
        ]

        if header.subtitle:
            parts.append(
                self._get_title(header.subtitle)[0],
            )

        parts.append(
            Part(
                text=None,
                commands=[
                    RepeatCommand(self.columns, "="),
                    NewLineCommand(),
                    AlignCommand(AlignCommand.Alignment.center)
                ],
            ),
        )

        return parts

    def _build_body(
        self,
        body,    # type: JustificationBodyDto
    ):
        # type: (...) -> List

        parts = []

        parts.extend(
            self._get_line(
                title="$ORDER",
                text=body.order_id,
                definition=self.body_definitions,
            )
        )

        parts.extend(
           self._get_line(
               title="$OPERATOR",
               text=body.operator,
               definition=self.body_definitions,
            )
        )

        parts.extend(
            self._get_line(
                title="$AUTHORIZED_BY",
                text=body.authorized_by,
                definition=self.body_definitions,
            )
        )

        parts.extend(
            [
                Part(
                    text=None,
                    commands=[
                        RepeatCommand(self.columns, "="),
                        AlignCommand(AlignCommand.Alignment.center),
                        NewLineCommand(),
                    ],
                )
            ]
        )

        if body.discount_description not in ['', None]:
            parts.extend(
                self._get_line(
                    title=None,
                    text="$DISCOUNT_JUSTIFICATION {0}".format(body.discount_description),
                    definition=self.center_definitions,
                )
            )

            parts.extend(
                [
                    Part(
                        text=None,
                        commands=[
                            NewLineCommand(),
                        ]
                    ),
                ]
            )

        parts.extend(
            self._get_line(
                title=None,
                text="$JUSTIFICATION {0}".format(body.justification),
                definition=self.center_definitions,
            )
        )

        return parts

    def _build_footer(
        self,
    ):
        # type: (...) -> List

        parts = [
            Part(
                text=None,
                commands=[
                    NewLineCommand(),
                    NewLineCommand(),
                    NewLineCommand(),
                    RepeatCommand(self.columns, "_"),
                    NewLineCommand(),
                    AlignCommand(AlignCommand.Alignment.center),
                ]
            ),
        ]

        parts.extend(
            self._get_line(
                title=None,
                text="$SIGNATURE",
                definition=self.footer_column_definition,
            )
        )

        return parts

    def _get_title(
        self,
        text,      # type: JustificationHeaderDto
    ):
        # type: (...) -> str

        return format_line(
            title=None,
            text=text,
            columns=self.columns,
            definition=self.header_column_definition,
        )

    def _get_line(
        self,
        title,              # type: str
        text,               # type: str
        definition,         # type: List[ReportColumnDefinition]
    ):
        # type: (...) -> List[TableReport]

        value = remove_accents(
           text=text,
        )

        return format_line(
            title=title,
            text=value,
            columns=self.columns,
            definition=definition,
        )



import json

from messageprocessorutil import UuidMessageProcessor
from messagebus import (
    NakMessage,
    AckMessage,
    MessageBus,
)

from production.interactor import PickListOrdersInteractor
from production.repository import ProductionRepository
from typing import (
    Dict,
    Optional,
)


class PickListOrdersProcessor(UuidMessageProcessor):

    def __init__(
        self,
        message_bus,                 # type: MessageBus
        production_repository,       # type: ProductionRepository
    ):
        # type: (...) -> None

        super(PickListOrdersProcessor, self).__init__()

        self.interactor = PickListOrdersInteractor(
            message_bus=message_bus,
            production_repository=production_repository,
        )

    def parse_data(
        self,
        data,        # type: str
    ):
        data = json.loads(
            s=data,
        )

        return data

    def call_business(
        self,
        obj,        # type: Dict
    ):
        # type: (...) -> Optional[Dict]

        return self.interactor.execute(
            obj=obj,
        )

    def format_response(self, message_bus, message, event, input_obj, result):
        message_bus.reply_message(
            message=message,
            reply_message=AckMessage(json.dumps(result)),
        )

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_message(
            message=message,
            reply_message=NakMessage(),
        )

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_message(
            message=message,
            reply_message=NakMessage(),
        )



def migrate(conn, tables):
    import os
    from systools import sys_log_debug
    updateTablesPriority = [
        "Orders", "OrderCustomProperties", "OrderItemCustomProperties", "OrderStateHistory",
        "OrderItem", "OrderTax", "OrderDiscount", "OrderVoidHistory", "OrderTender", "OrderComment",
        "POS", "CurrentOrderItem", "BackupOrders", "BackupCurrentOrderItem", "OrderPrice", "OrderProduct"
    ]
    sys_log_debug("[migrate] start")
    for table in updateTablesPriority:
        if tables[table]["fields_OldDB"]:
            sys_log_debug("[migrate] table {}".format(table))
            strCommonFields = ", ".join(tables[table]["fields_Common"])
            sys_log_debug("[migrate] copying data for table {}".format(table))
            stmt = "INSERT OR REPLACE INTO %s (%s) SELECT %s FROM old.%s;" % (table, strCommonFields, strCommonFields, table)
            conn.query(stmt)
            sys_log_debug("[migrate] finished query for table {}".format(table))
    # Updates new OrderPrice table
    if not tables["OrderPrice"]["fields_OldDB"]:
        db_path = '{}/databases/product.db'.format(os.environ['HVDATADIR'])
        sys_log_debug("[migrate] attaching database {}".format(db_path))
        conn.query("ATTACH DATABASE '{}' as 'productdb';".format(db_path))
        sys_log_debug("[migrate] copying actual OrderItem pricekeys to new OrderPrice table ...")
        conn.query("""
            INSERT INTO OrderPrice(OrderId,PriceKey,DefaultUnitPrice,AddedUnitPrice,SubtractedUnitPrice,Computed,IncludedQty,ValidFrom,ValidThru)
            SELECT T.OrderId,T.PriceKey,P.DefaultUnitPrice,P.AddedUnitPrice,P.SubtractedUnitPrice,P.Computed,P.IncludedQty,P.ValidFrom,P.ValidThru
            FROM (SELECT DISTINCT OrderId, PriceKey FROM OrderItem) T
            JOIN productdb.Price P WHERE P.PriceKey=T.PriceKey;
        """)
    sys_log_debug("[migrate] finished iterating")



from ._AutoMappedMessageProcessor import AutoMappedMessageProcessor


def auto_map(params):
    def inner_auto_map(cls):
        class NewCls(object):
            _cls_attrs = params
            _real_cls = cls

            def __init__(self, *args, **kwargs):
                self.oInstance = cls(*args, **kwargs)

            def __getattribute__(self, s):
                try:
                    x = super(NewCls, self).__getattribute__(s)
                except AttributeError:
                    pass
                else:
                    return x
                return self.oInstance.__getattribute__(s)

        return NewCls

    return inner_auto_map


def exec_map(params):
    def inner_exec_map(cls):
        class NewCls(Exception):
            _cls_attrs = params
            _real_cls = cls

            def __init__(self, *args, **kwargs):
                self.oInstance = cls(*args, **kwargs)

            def __getattribute__(self, s):
                try:
                    x = super(NewCls, self).__getattribute__(s)
                except AttributeError:
                    pass
                else:
                    return x
                return self.oInstance.__getattribute__(s)

        return NewCls

    return inner_exec_map



from application.service import OrderService, BenefitService
from application.model import Configurations
from typing import Dict, List, Optional


class OnOrderReopenInteractor(object):

    def __init__(self, configs, order_service, benefit_service):
        # type: (Configurations, OrderService, BenefitService) -> None

        self.configs = configs
        self.order_service = order_service
        self.benefit_service = benefit_service

    def execute(self, pos_id):
        # type: (int) -> None

        order_picture = self.order_service.get_order_picture(pos_id)

        state = order_picture.get("state", "") if order_picture else ""
        if state != "IN_PROGRESS":
            return

        order = self.order_service.get_order(pos_id, order_picture=order_picture)
        order_picture = order.order_picture
        remove_manual = order.table_id is None

        applied_benefits = self.order_service.get_applied_benefits_from_order_picture(
            pos_id=pos_id,
            order_picture=order_picture,
            return_benefit_pre_total_sequence=True,
        )
        if not applied_benefits:
            return

        benefits_to_remove = self._get_benefits_to_remove(applied_benefits)
        if len(benefits_to_remove) > 0:
            self.benefit_service.remove_benefits(pos_id, benefits_to_remove, order_picture, delete_manual=remove_manual)

        self.benefit_service.unset_applied_benefit(pos_id, ["IN_PROGRESS"])

    def _get_benefits_to_remove(self, applied_benefits):
        # type: (List) -> List

        benefits_to_remove = []
        for applied_benefit in applied_benefits["benefitList"]:
            benefit_key_to_remove = self._get_benefit_id_to_remove(
                applied_benefit=applied_benefit,
                benefit_pre_total_sequence=applied_benefits["benefitPreTotalSequence"]
            )
            if benefit_key_to_remove:
                benefits_to_remove.append(benefit_key_to_remove)

        return benefits_to_remove

    @staticmethod
    def _get_benefit_id_to_remove(
        applied_benefit,                # type: Dict
        benefit_pre_total_sequence,     # type: Dict
    ):
        # type: (...) -> Optional[int]

        benefit_applied = applied_benefit.get("benefit_applied")
        if not benefit_applied:
            return None

        benefit = applied_benefit.get("benefit", {})
        is_loyalty_benefit = benefit.get("is_loyalty_benefit", False)
        if (
            applied_benefit.get("benefit_key", "") in benefit_pre_total_sequence
            and benefit.get("automatic", "") != "ON_TOTAL"
            and not is_loyalty_benefit
        ):
            return None

        return applied_benefit.get("benefit_key")



import logging
from typing import Any, Optional

from flask import Flask, request, Response
from httpclient import HttpRequest
from httpmessagehandler import HttpMessageHandler

logger = logging.getLogger()


class FlaskMessageHandler(object):
    def __init__(self, http_message_handler):
        # type: (HttpMessageHandler) -> None
        self.http_message_handler = http_message_handler

        self.app = None

    def run(self, name, host=None, port=None):
        # type: (str, Optional[str], Optional[int]) -> Any
        self.http_message_handler.build_singletons()

        self.app = Flask(name)
        self.app.add_url_rule("/", "catch_all",
                              self.handle_messages,
                              defaults={"path": "/"},
                              methods=["GET", "POST", "PUT", "OPTIONS", "HEAD", "DELETE", "PATCH"])
        self.app.add_url_rule("/<path:path>",
                              "catch_all",
                              self.handle_messages,
                              methods=["GET", "POST", "PUT", "OPTIONS", "HEAD", "DELETE", "PATCH"])
        if host is not None:
            self.app.run(host=host, port=port)
            self.http_message_handler.destroy_singletons()

        return self.app

    def handle_messages(self, path):
        # type: (str) -> Response
        if path != "/":
            path = "/" + path

        http_request = HttpRequest(path, request.method, request.args, self._get_headers(), request.get_data())
        http_response = self.http_message_handler.handle_request(http_request)
        return Response(status=http_response.status_code,
                        response=http_response.body,
                        headers=http_response.headers)

    @staticmethod
    def _get_headers():
        headers = {}
        if request.headers is not None:
            for key in request.headers.keys():
                headers[key] = request.headers[key]
        return headers




import json

from old_helper import BaseRepository
from persistence import Connection
from typing import List

from justificationdiscountreport.model import JustificationDiscount


class JustificationDiscountRepository(BaseRepository):

    def __init__(
        self,
        mb_context,
        pos_id,
    ):
        # type: (...) -> None

        super(JustificationDiscountRepository, self).__init__(mb_context)

        self.pos_id = pos_id

    def get_justifications(
        self,
        initial_date,       # type: str
        end_date,           # type: str
        operator_id,        # type: str
    ):
        # type: (...) -> List[JustificationDiscount]

        def inner_func(
            conn,    # type: Connection
        ):
            # type: (...) -> List[JustificationDiscount]

            query = self._OrdersQuery.format(initial_date, end_date)
            result = []
            for row in conn.select(query):
                benefit_list_json = json.loads(row.get_entry("BenefitList"))
                benefit_list = json.loads(benefit_list_json.get("benefitList", None))
                for benefit_item in benefit_list:

                    benefit = benefit_item.get("benefit")
                    request_justification = benefit.get("requestJustification")
                    justification = benefit.get("justification", None)

                    if not request_justification or justification in ['', None]:
                        continue

                    applied_by = benefit.get("appliedBy", "")

                    if operator_id is not None and operator_id != applied_by:
                        continue

                    benefit_id = benefit_item.get("benefit_id", "")
                    benefit_total_value = benefit_item.get("benefit_total_value", 0)

                    result.append(
                        JustificationDiscount(
                            row_database=row,
                            benefit_id=benefit_id,
                            justification=justification,
                            applied_by=applied_by,
                            benefit_total_value=benefit_total_value,
                        )
                    )

            return result

        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, self.pos_id)

    _OrdersQuery = """
                        SELECT o.OrderId,
                               o.TotalGross,
                               od.OrderDate,
                               od.OrderTime,
                               ocp.Value AS BenefitList
                         FROM Orders o
                         INNER JOIN OrderCustomProperties ocp ON ocp.OrderId = o.OrderId AND ocp.Key = 'BENEFIT_LIST'
                         INNER JOIN (
                                    SELECT OrderId, 
                                           strftime('%Y-%m-%d', max(Timestamp) , 'localtime') OrderDate,
                                           strftime('%H:%M:%S', max(Timestamp), 'localtime') OrderTime
                                      FROM OrderStateHistory
                                     WHERE StateId = 5
                                     GROUP BY OrderId
                                ) od on o.OrderId = od.OrderId
                         WHERE 
                            o.BusinessPeriod >= '{0}' 
                            AND o.BusinessPeriod <= '{1}' 
                            AND o.StateId = 5
                         GROUP by o.OrderId, o.BusinessPeriod
                         ORDER BY o.OrderId, o.BusinessPeriod
                      """



# -*- coding: utf-8 -*-
"""
    flask.__main__
    ~~~~~~~~~~~~~~

    Alias for flask.run for the command line.

    :copyright: 2010 Pallets
    :license: BSD-3-Clause
"""

if __name__ == "__main__":
    from .cli import main

    main(as_module=True)



class Config(object):

    def __init__(self):
        self.loader_path = None
        self.component_name = None
        self.output_folder = None
        self.encode = None
        self.log_level = None
        self.connection_timeout = None



from _JustificationDiscountReportGenerator import JustificationDiscountReportGenerator
from _JustificationDiscountReportPresenter import JustificationDiscountReportPresenter

__all__ = [
    "JustificationDiscountReportGenerator",
    "JustificationDiscountReportPresenter",
]



from production.model import Item
from typing import Any

from ._AllowedForbiddenFilter import AllowedForbiddenFilter
from ._OrderChangerProductionBox import OrderChangerProductionBox


def item_properties_retriever(item):
    # type: (Item) -> Any
    ret = []
    for p in item.properties:
        ret.append(p)

    return ret


class ItemPropertiesFilterBox(OrderChangerProductionBox):
    def __init__(self, name, sons, allowed_properties, forbidden_properties, allow_combos, logger=None):
        super(ItemPropertiesFilterBox, self).__init__(name, sons, logger)
        self.allowed_forbidden_filter = AllowedForbiddenFilter(allowed_properties, forbidden_properties, allow_combos)

    def change_order(self, order):
        return self.allowed_forbidden_filter.filter_items(order, item_properties_retriever)



from _OperatorLogoutInteractor import OperatorLogoutInteractor
from _OrderPictureInteractor import OrderPictureInteractor
from _TablePictureInteractor import TablePictureInteractor
from _ProductionOrderInteractor import ProductionOrderInteractor
from _StoreStatusInteractor import StoreStatusInteractor



import json
import logging

from msgbus import (
    FM_PARAM,
    TK_SYS_ACK, MBException,
)
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
)

from mw_helper import (
    show_check_list_box_dialog,
    show_filtered_list_box_dialog,
    show_message_dialog,
    show_message_options_dialog,
    show_loyalty_dialog,
)
from sysactions import (
    action,
    get_model,
    get_posot,
    send_message,
    show_messagebox,
    translate_message,
)

from bustoken import TK_LOYALTY_GET_LOYALTY_ID

from actions.customer import get_customer_doc
from actions.loyalty.model import LoyaltyButtonEnum
from utilfunctions import (
    get_custom_property_value,
    parse_loyalty_customer_info,
    update_custom_properties,
)


@action
def open_loyalty(
    pos_id,                 # type: str
    table_id=None,          # type: Optional[str]
    order_id=None,          # type: Optional[str]
    table_seat="table",     # type: Optional[str]
):
    # type: (...) -> bool

    while True:
        loyalty_properties = _get_loyalty_properties(
                pos_id=pos_id,
                table_id=table_id,
                order_id=order_id,
            )

        if loyalty_properties:
            has_customer_info = _has_customer_info(
                loyalty_properties=loyalty_properties,
                table_id=table_id,
                order_id=order_id,
                table_seat=table_seat,
            )
        else:
            has_customer_info = False

        if has_customer_info:
            welcome_message = _get_welcome_message(
                loyalty_properties=loyalty_properties,
                table_id=table_id,
                order_id=order_id,
                table_seat=table_seat,
            )
            cpf_button = "$REMOVE_CPF"
            is_disabled = ""
        else:
            welcome_message = "$INSERT_CUSTOMER_DOC"
            cpf_button = "$INFORM_CPF"
            is_disabled = '#DISABLED'

        buttons = "{}|$AVAILABLE_COUPONS{}|$APPLIED_COUPONS{}|$CLOSE".format(
            cpf_button,
            is_disabled,
            is_disabled,
        )
        resp = show_loyalty_dialog(
            pos_id=pos_id,
            buttons=buttons,
            message=welcome_message,
        )

        if resp in (None, LoyaltyButtonEnum.CANCEL):
            break

        if resp == LoyaltyButtonEnum.UPDATE_CPF:
            if has_customer_info:
                _remove_loyalty_properties(
                    pos_id=pos_id,
                    table_id=table_id,
                    order_id=order_id,
                    table_seat=table_seat,
                )
            else:
                _set_loyalty_properties(
                    pos_id=pos_id,
                    table_id=table_id,
                    order_id=order_id,
                    table_seat=table_seat,
                )

        elif resp == LoyaltyButtonEnum.AVAILABLE_COUPONS:
            if has_customer_info:
                _show_available_coupons(
                    pos_id=pos_id,
                    table_id=table_id,
                    order_id=order_id,
                    table_seat=table_seat,
                )

        elif resp == LoyaltyButtonEnum.APPLIED_COUPONS and has_customer_info:
            _show_applied_coupons(
                pos_id=pos_id,
                table_id=table_id,
                order_id=order_id,
                table_seat=table_seat,
            )

    return True


def _get_table_customer_doc(
        pos_id,         # type: str
        table_id,       # type: str
        table_seat,     # type: str
):
    # type: (...) -> Optional[str]

    from msgbusboundary import MsgBusTableService as tableService

    customer_doc = tableService().get_table_custom_property(
        pos_id=pos_id,
        prop="LOYALTY_TABLE_CUSTOMER_INFO",
        table_id=table_id,
    )
    if customer_doc:
        customer_doc = json.loads(s=customer_doc['LOYALTY_TABLE_CUSTOMER_INFO'])
        if customer_doc.get(table_seat, {}):
            customer_doc = customer_doc[table_seat].get('customer_doc', None)
        else:
            customer_doc = None

    return customer_doc


def get_loyalty_info(
    customer_doc,   # type: str
):
    # type: (...) -> Optional[str]

    if not customer_doc:
        return

    try:
        ret = send_message(
            dest_name="Loyalty",
            token=TK_LOYALTY_GET_LOYALTY_ID,
            format=FM_PARAM,
            data=customer_doc,
        )
        if ret.token == TK_SYS_ACK:
            return ret.data
    except MBException as ex:
        if "SE_NOTFOUND" in ex.message:
            return None

        raise ex


def _set_loyalty_properties(
    pos_id,          # type: str
    table_id,        # type: str
    order_id,        # type: str
    table_seat,      # type: str
):
    # type: (...) -> None

    if order_id not in [None, ""]:
        from posactions import doSetCustomerDocument
        doSetCustomerDocument(
            pos_id=pos_id,
            stop_action=False,
        )

    elif table_id is not None:
        loyalty_properties = _build_loyalty_properties(
            pos_id=pos_id,
            table_id=table_id,
            table_seat=table_seat,
        )
        _persist_loyalty_properties_ts(
            loyalty_properties=loyalty_properties,
            pos_id=pos_id,
            table_id=table_id,
            table_seat=table_seat,
        )


def _persist_loyalty_properties_ts(
    loyalty_properties,             # type: Dict
    pos_id,                         # type: str
    table_id,                       # type: str
    table_seat,                     # type: str
    only_discounts_on_total=False,  # type: bool
):
    # type: (...) -> None

    from msgbusboundary import MsgBusTableService as tableService
    from table_actions import create_table_modified_event

    customer_info = {}
    if not only_discounts_on_total:
        customer_info = loyalty_properties.get("LOYALTY_TABLE_CUSTOMER_INFO")
        customer_info_value = json.dumps(customer_info) if customer_info else "{}"
        tableService().set_table_custom_property(
            pos_id=pos_id,
            prop="LOYALTY_TABLE_CUSTOMER_INFO",
            prop_value=customer_info_value,
            table_id=table_id,
        )

    if not only_discounts_on_total:
        benefits = loyalty_properties.get("LOYALTY_TABLE_CUSTOMER_BENEFITS")
        benefits_value = json.dumps(benefits) if benefits else "{}"
        tableService().set_table_custom_property(
            pos_id=pos_id,
            prop="LOYALTY_TABLE_CUSTOMER_BENEFITS",
            prop_value=benefits_value,
            table_id=table_id,
        )

    coupons_on_total = loyalty_properties.get("LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL")
    coupons_on_total_value = json.dumps(coupons_on_total) if coupons_on_total else "{}"
    tableService().set_table_custom_property(
        pos_id=pos_id,
        prop="LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL",
        prop_value=coupons_on_total_value,
        table_id=table_id,
    )

    if not only_discounts_on_total and (table_seat == "table" or _is_tab(table_id=table_id)):
        table_customer = customer_info.get("table", {})
        if isinstance(table_customer, basestring):
            table_customer = json.loads(table_customer)
        customer_info = table_customer.get("customerInfo", {})
        customer = customer_info.get("customer", {})
        document = customer.get("document", "")
        name = customer.get("name", "")
        tableService().set_table_custom_property(
            pos_id=pos_id,
            prop="CUSTOMER_DOC",
            prop_value=document,
            table_id=table_id,
        )
        tableService().set_table_custom_property(
            pos_id=pos_id,
            prop="CUSTOMER_NAME",
            prop_value=name,
            table_id=table_id,
        )

    table_picture = tableService().get_table_picture(
        pos_id=pos_id,
        table_id=table_id,
    )

    create_table_modified_event(
        pos_id=pos_id,
        table_picture=table_picture,
    )


def _get_loyalty_properties(
    pos_id,         # type: str
    table_id=None,  # type: Optional[str]
    order_id=None,  # type: Optional[str]
):
    # type: (...) -> Optional[Dict]

    if order_id:
        return _get_loyalty_properties_for_qsr(
            pos_id=pos_id,
            order_id=order_id,
        )

    if table_id:
        return _get_loyalty_properties_for_ts(
            pos_id=pos_id,
            table_id=table_id,
        )

    return None


def _get_loyalty_properties_for_qsr(
    pos_id,    # type: str
    order_id,  # type: str
):
    # type: (...) -> Optional[Dict]

    customer_info = get_custom_property_value(
        pos_id=pos_id,
        custom_property="LOYALTY_CUSTOMER_INFO",
        order_id=order_id,
    )
    benefits = get_custom_property_value(
        pos_id=pos_id,
        custom_property="LOYALTY_CUSTOMER_BENEFITS",
        order_id=order_id,
    )
    coupons_on_total = get_custom_property_value(
        pos_id=pos_id,
        custom_property="LOYALTY_COUPONS_TO_APPLY_ON_TOTAL",
        order_id=order_id,
    )
    customer_info = json.loads(customer_info) if customer_info else []
    benefits = json.loads(benefits) if benefits else []
    coupons_on_total = json.loads(coupons_on_total) if coupons_on_total else []

    loyalty_properties = {
        "LOYALTY_CUSTOMER_INFO": customer_info,
        "LOYALTY_CUSTOMER_BENEFITS": benefits,
        "LOYALTY_COUPONS_TO_APPLY_ON_TOTAL": coupons_on_total,
    }
    return loyalty_properties


def _get_loyalty_properties_for_ts(
    pos_id,     # type: str
    table_id,   # type: str
):
    # type: (...) -> Optional[Dict]

    from msgbusboundary import MsgBusTableService as tableService

    customer_info = tableService().get_table_custom_property(
        pos_id=pos_id,
        prop="LOYALTY_TABLE_CUSTOMER_INFO",
        table_id=table_id
    )
    benefits = tableService().get_table_custom_property(
        pos_id=pos_id,
        prop="LOYALTY_TABLE_CUSTOMER_BENEFITS",
        table_id=table_id
    )
    coupons_on_total = tableService().get_table_custom_property(
        pos_id=pos_id,
        prop="LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL",
        table_id=table_id
    )

    if customer_info:
        customer_info = json.loads(customer_info["LOYALTY_TABLE_CUSTOMER_INFO"])
        _parse_object_children(dictionary=customer_info)

    if benefits:
        benefits = json.loads(benefits["LOYALTY_TABLE_CUSTOMER_BENEFITS"])
        _parse_object_children(dictionary=benefits)

    if coupons_on_total:
        coupons_on_total = json.loads(coupons_on_total["LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL"])
        _parse_object_children(dictionary=coupons_on_total)

    loyalty_properties = {
        "LOYALTY_TABLE_CUSTOMER_INFO": customer_info,
        "LOYALTY_TABLE_CUSTOMER_BENEFITS": benefits,
        "LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL": coupons_on_total,
    }
    return loyalty_properties


def _parse_object_children(
    dictionary,     # type: Dict
):
    # type: (...) -> None

    for key, value in dictionary.items():
        if isinstance(value, basestring):
            dictionary.update({key: json.loads(value)})


def _remove_loyalty_properties(
    pos_id,        # type: str
    table_id,      # type: str
    order_id,      # type: str
    table_seat,    # type: str
):
    # type: (...) -> None

    if order_id not in [None, ""]:
        model = get_model(posid=pos_id)
        pos_ot = get_posot(model=model)
        order_properties_dict = update_custom_properties(
            customer_doc="clear",
            customer_name="clear",
            loyalty_customer_info="clear",
        )

        for prop in order_properties_dict.keys():
            if order_properties_dict[prop] in [None, ""]:
                pos_ot.setOrderCustomProperty(key=prop)
                del order_properties_dict[prop]

        if len(order_properties_dict) > 0:
            pos_ot.setOrderCustomProperties(properties=order_properties_dict)

    elif table_id is not None:
        loyalty_properties = _get_loyalty_properties(
            pos_id=pos_id,
            table_id=table_id,
            order_id=order_id,
        )

        for property_name, value in loyalty_properties.items():
            if table_seat in value:
                loyalty_properties[property_name].pop(table_seat)

        _persist_loyalty_properties_ts(
            loyalty_properties=loyalty_properties,
            pos_id=pos_id,
            table_id=table_id,
            table_seat=table_seat
        )


def _build_loyalty_properties(
    pos_id,                 # type: str
    table_id=None,          # type: Optional[str]
    order_id=None,          # type: Optional[str]
    table_seat='table',     # type: Optional[str]
):
    # type: (...) -> Dict

    loyalty_properties = _get_loyalty_properties(
        pos_id=pos_id,
        table_id=table_id,
        order_id=order_id,
    )

    model = get_model(posid=pos_id)
    pos_ot = get_posot(model=model)
    document = get_customer_doc(
        pos_id=pos_id,
        pos_ot=pos_ot,
        stop_action=False,
    )
    if not document:
        return loyalty_properties

    loyalty_customer_info = get_loyalty_info(customer_doc=document)
    result = parse_loyalty_customer_info(loyalty_customer_info=loyalty_customer_info)

    if not result:
        return loyalty_properties

    if order_id:
        loyalty_properties = result

    elif table_id:
        result = {
            key.replace("LOYALTY_", "LOYALTY_TABLE_"): value
            for key, value in result.items()
        }
        for key, value in loyalty_properties.items():
            if key not in result:
                continue
            value.update({table_seat: result[key]})

    return loyalty_properties


def _has_customer_info(
    loyalty_properties,     # type: Dict
    table_id=None,          # type: Optional[str]
    order_id=None,          # type: Optional[str]
    table_seat="table",     # type: Optional[str]
):
    # type: (...) -> bool

    if order_id:
        customer_info = loyalty_properties.get("LOYALTY_CUSTOMER_INFO", {})
        return bool(customer_info)

    if table_id:
        customer_info = loyalty_properties.get("LOYALTY_TABLE_CUSTOMER_INFO", {})
        customer_info = customer_info.get(table_seat)
        return bool(customer_info)

    return False


def _get_welcome_message(
    loyalty_properties,     # type: Dict
    table_id=None,          # type: Optional[str]
    order_id=None,          # type: Optional[str]
    table_seat="table",     # type: Optional[str]
):
    # type: (...) -> unicode

    if order_id:
        customer_info = loyalty_properties.get("LOYALTY_CUSTOMER_INFO", {})

    elif table_id:
        customer_info = loyalty_properties.get("LOYALTY_TABLE_CUSTOMER_INFO", {})
        customer_info = customer_info.get(table_seat, {})

    else:
        return "$LOYALTY_WELCOME_MESSAGE_CUSTOMER"

    customer_info = customer_info.get("customerInfo", {})
    customer = customer_info.get("customer", {})
    customer_name = customer.get("name", "")
    customer_doc = customer.get("document", "")

    additional_info = customer_info.get("additionalInfo", {})
    welcome_message = additional_info.get("welcomeMessage", "")

    message = "$WELCOME_CLIENT_LOYALTY_INFO|{}|{}|{}".format(
        customer_name,
        customer_doc,
        welcome_message.encode('utf-8'),
    )
    return message


def _show_available_coupons(
    pos_id,              # type: str
    table_id=None,       # type: Optional[str]
    order_id=None,       # type: Optional[str]
    table_seat="table",  # type: Optional[str]
):
    # type: (...) -> None

    if not (order_id or table_id):
        return

    loyalty_properties = _get_loyalty_properties(
        pos_id=pos_id,
        table_id=table_id,
        order_id=order_id,
    )

    if not loyalty_properties:
        return

    customer_benefits, coupons_to_apply_on_total, _ = _get_loyalty_coupons(
        loyalty_properties=loyalty_properties,
        order_id=order_id,
        table_id=table_id,
        table_seat=table_seat,
    )

    customer_benefits = customer_benefits or []
    coupons_to_apply_on_total = coupons_to_apply_on_total or []

    available_coupons = [
        {
            "coupon": benefit,
            "name": benefit.get("userDescription", ""),
            "benefitId": benefit.get("id"),
        }
        for index, benefit in enumerate(customer_benefits)
        if benefit.get("id") not in coupons_to_apply_on_total
    ]

    if not available_coupons:
        show_message_dialog(
            pos_id=pos_id,
            title="$INFORMATION",
            message="$WITHOUT_ANY_LOYALTY_BENEFITS_TO_SHOW"
        )
        return

    options = [
        "{} - {}".format(
            index + 1,
            coupon["name"].encode("utf-8")
        )
        for index, coupon in enumerate(available_coupons)
    ]

    selected_index = show_filtered_list_box_dialog(pos_id, options, "$SELECT_ONE_COUPON", mask="NOFILTER")

    if selected_index is None:
        return
    else:
        benefit_id = available_coupons[selected_index]["benefitId"]
        _apply_selected_benefit(
            loyalty_properties=loyalty_properties,
            pos_id=pos_id,
            benefit_id=benefit_id,
            table_id=table_id,
            order_id=order_id,
            table_seat=table_seat,
        )


def _get_loyalty_coupons(
    loyalty_properties,      # type: Dict
    order_id=None,           # type: Optional[str]
    table_id=None,           # type: Optional[str]
    table_seat="table",      # type: Optional[str]
    current_used_key=False,  # type: Optional[bool]
):
    # type: (...) -> Tuple[Dict, List, str]

    customer_benefits = None
    coupons_to_apply_on_total = None

    if order_id:
        customer_benefits = loyalty_properties["LOYALTY_CUSTOMER_BENEFITS"]
        coupons_to_apply_on_total = loyalty_properties["LOYALTY_COUPONS_TO_APPLY_ON_TOTAL"]
        current_used_key = "LOYALTY_COUPONS_TO_APPLY_ON_TOTAL"

    elif table_id:
        customer_benefits = loyalty_properties["LOYALTY_TABLE_CUSTOMER_BENEFITS"]
        customer_benefits = customer_benefits.get(table_seat)
        coupons_to_apply_on_total = loyalty_properties["LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL"]
        coupons_to_apply_on_total = coupons_to_apply_on_total.get(table_seat)
        current_used_key = "LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL"

    return customer_benefits, coupons_to_apply_on_total, current_used_key


def _show_applied_coupons(
    pos_id,                # type: str
    table_id=None,         # type: Optional[str]
    order_id=None,         # type: Optional[str]
    table_seat="table",    # type: Optional[str]
):
    # type: (...) -> None

    loyalty_properties = _get_loyalty_properties(
        pos_id=pos_id,
        table_id=table_id,
        order_id=order_id,
    )

    customer_benefits, coupons_to_apply_on_total, current_used_key = _get_loyalty_coupons(
        loyalty_properties=loyalty_properties,
        order_id=order_id,
        table_id=table_id,
        table_seat=table_seat,
        current_used_key=True,
    )

    if not customer_benefits or not coupons_to_apply_on_total:
        show_messagebox(
            posid=pos_id,
            message="$ORDER_HAS_NO_APPLIED_DISCOUNT",
        )
        return

    coupon_options = _format_coupon_options(
        pos_id=pos_id,
        coupon_options=customer_benefits,
        coupons_to_apply_on_total=coupons_to_apply_on_total,
    )
    options = ["{0}".format(c["name"]) for c in coupon_options]
    selected_indexes = show_check_list_box_dialog(
        pos_id=pos_id,
        options=options,
        title="$SELECT_ONE_COUPON",
        buttons="$REMOVE|$CANCEL",
        mask="NO_SALE_PANEL_IN_LOYALTY_BENEFITS",
    )

    if not len(selected_indexes):
        return

    if len(selected_indexes) == len(coupon_options):
        if order_id:
            loyalty_properties[current_used_key] = []
        elif table_id:
            loyalty_properties[current_used_key][table_seat] = []

    else:
        loyalty_property = []
        if order_id:
            loyalty_property = loyalty_properties[current_used_key]
        elif table_id:
            loyalty_property = loyalty_properties[current_used_key][table_seat]

        for selected_index in selected_indexes:
            index = int(selected_index)
            coupon_item = [x for x in coupon_options if x["index"] == index]
            coupon_id = coupon_item[0]['id']

            if index == 0:
                continue
            if not coupon_item:
                continue

            loyalty_property.remove(coupon_id)

    _apply_selected_benefit(
        loyalty_properties=loyalty_properties,
        pos_id=pos_id,
        benefit_id=None,
        table_id=table_id,
        order_id=order_id,
        table_seat=table_seat,
    )


def _format_coupon_options(
    pos_id,                       # type: str
    coupon_options,               # type: Dict
    coupons_to_apply_on_total,    # type: List
):
    # type: (...) -> List

    model = get_model(pos_id)
    all_coupons = translate_message(model, "ALL")
    applied_coupons = []
    benefit_index = 1

    for coupon in coupon_options:
        benefit_id = coupon.get('id')

        if benefit_id not in coupons_to_apply_on_total:
            continue

        disable = False
        description = (coupon.get("userDescription") or coupon.get("id") or "").encode("UTF-8")
        name = "{}##{}##{}".format(description, 'null', disable)
        data = {"index": benefit_index, "id": benefit_id, "name": name}
        applied_coupons.append(data)
        benefit_index += 1

    name = "{}##{}##False".format(all_coupons, 'null')
    data = {"index": 0, "id": None, "name": name}
    applied_coupons.append(data)
    applied_coupons.sort(key=lambda x: (x["index"]))
    return applied_coupons


def _apply_selected_benefit(
    loyalty_properties,  # type: Dict
    pos_id,              # type: str
    benefit_id,          # type: Optional[str]
    table_id=None,       # type: Optional[str]
    order_id=None,       # type: Optional[str]
    table_seat="table",  # type: Optional[str]
):
    # type: (...) -> None

    if order_id:
        coupons_to_apply_on_total = loyalty_properties["LOYALTY_COUPONS_TO_APPLY_ON_TOTAL"]
        if benefit_id:
            coupons_to_apply_on_total.append(benefit_id)

        model = get_model(posid=pos_id)
        pos_ot = get_posot(model=model)
        pos_ot.setOrderCustomProperty(
            "LOYALTY_COUPONS_TO_APPLY_ON_TOTAL",
            value=json.dumps(coupons_to_apply_on_total),
            orderid=order_id
        )

    elif table_id:
        coupons_to_apply_on_total = loyalty_properties["LOYALTY_TABLE_COUPONS_TO_APPLY_ON_TOTAL"]
        if benefit_id:
            if table_seat in coupons_to_apply_on_total:
                coupons_to_apply_on_total[table_seat].append(benefit_id)
            else:
                coupons_to_apply_on_total[table_seat] = [benefit_id]

        _persist_loyalty_properties_ts(
            loyalty_properties=loyalty_properties,
            pos_id=pos_id,
            table_id=table_id,
            table_seat=table_seat,
            only_discounts_on_total=True,
        )


def _is_tab(
    table_id,   # str
):
    # type: (...) -> bool

    return 'TAB' in table_id



from logging import Logger

from messagebus import MessageBus
from production.manager import ProductionManager
from production.model import BuilderNotFoundException
from production.treebuilder import ViewConfiguration
from production.treebuilder.viewfactory.builder import KdsViewBuilder, PublishViewBuilder, CommandViewBuilder
from production.view import ProductionView
from typing import Dict

from .builder import PrinterViewBuilder


class ViewFactory(object):
    def __init__(self, message_bus, loggers, production_manager, kds_manager):
        # type: (MessageBus, Dict[str, Logger], ProductionManager, str) -> None

        self.builders = {
            u"PrinterView": PrinterViewBuilder(
                message_bus=message_bus,
                loggers=loggers,
                production_manager=production_manager,
            ),
            u"KdsView": KdsViewBuilder(message_bus, kds_manager, loggers),
            u"PublishView": PublishViewBuilder(message_bus, loggers),
            u"CommandView": CommandViewBuilder(production_manager, loggers)
        }

    def build(self, config):
        # type: (ViewConfiguration) -> ProductionView
        if config.type not in self.builders:
            raise BuilderNotFoundException("Builder not found for view: {} - {}".format(config.name, config.type))

        return self.builders[config.type].build(config)



from msgbus import MBEasyContext, MBMessage, TK_SYS_NAK, TK_SYS_ACK, FM_STRING, FM_PARAM
from messagehandler import EventHandler
import logging

from taxcalculator import TaxCalculatorService

logger = logging.getLogger("NTaxCalc")


class NTaxCalcEventHandler(EventHandler):
    def __init__(self, mbcontext, tax_calculator_service):
        # type: (MBEasyContext, TaxCalculatorService) -> None
        super(NTaxCalcEventHandler, self).__init__(mbcontext)
        self.tax_calculator_service = tax_calculator_service

    def get_handled_tokens(self):
        return []

    def handle_event(self, subject, evt_type, data, msg):
        # type: (unicode, unicode, str, MBMessage) -> None

        try:
            if subject == "TAX_CALC" or (subject == "TAXCALC" and evt_type == "TAXCALC"):
                response = self.tax_calculator_service.calculate_taxes(data)
                msg.data = response
                msg.token = TK_SYS_ACK

                if evt_type == "TAXCALC":
                    response = '0\x00'+response
                self.mbcontext.MB_ReplyMessage(msg, data=response.encode("utf-8"), format=FM_PARAM)

            elif subject == "TAXCALC" and evt_type == "PRODTXIDX":
                response = self.tax_calculator_service.insert_tx_idx(data)
                msg.data = response
                msg.token = TK_SYS_ACK
                self.mbcontext.MB_ReplyMessage(msg, data=response.encode("utf-8"), format=FM_STRING)

            elif subject == "TAX_CALC_ITEM":
                response = self.tax_calculator_service.calculate_taxes_item(data)
                msg.data = response
                msg.token = TK_SYS_ACK
                self.mbcontext.MB_ReplyMessage(msg, data=response.encode("utf-8"), format=FM_STRING)
            else:
                msg.token = TK_SYS_ACK
                self.mbcontext.MB_EasyReplyMessage(msg)
        except Exception as _:  # noqa
            logger.exception("Erro tratando evento: {} / Data: {}".format(subject, data))

            msg.token = TK_SYS_NAK
            self.mbcontext.MB_ReplyMessage(msg)

    def handle_message(self, msg):
        return

    def terminate_event(self):
        return



import base64
import cPickle as Pickle
from Queue import Queue, Empty
from datetime import datetime
from threading import Thread

import production.repository
from mw_helper import BaseRepository
from persistence import Connection
from production.model import ProductionOrder
from typing import (
    List,
    Dict,
)


class ProductionRepository(BaseRepository, production.repository.ProductionRepository):
    def __init__(self, mb_context, logger):
        super(ProductionRepository, self).__init__(mb_context)
        self.logger = logger

        self.order_queue = Queue()
        self.running = True

        self.processor_thread = Thread(target=self.consume_queue)
        self.processor_thread.daemon = True
        self.processor_thread.start()

    def consume_queue(self):
        while self.running:
            order_id = None
            try:
                order_to_save = self.order_queue.get(True, 5)
                order_id = order_to_save.order_id
                self.save_order_async(order_to_save)
            except Empty:
                continue
            except BaseException as _:  # noqa
                self.logger.exception("Error saving order: {}".format(order_id))

    def stop(self):
        self.running = False

    def save_order(self, order):
        self.order_queue.put(order)

    def save_order_async(self, order):
        def func(conn):
            # type: (Connection) -> None
            order_data = base64.b64encode(Pickle.dumps(order, Pickle.HIGHEST_PROTOCOL))
            conn.query(insert_query.format(order.order_id, order_data, order.created_at, 0))

        self.execute_with_connection(func)

    def get_all_orders(self):
        def func(conn):
            # type: (Connection) -> List[ProductionOrder]
            orders = []
            for row in conn.select(select_query):
                order_data = base64.b64decode(row.get_entry(0))
                order = Pickle.loads(order_data)
                orders.append(order)

            return orders

        return self.execute_with_connection(func)

    def get_orders_to_purge(self, limit_date):
        def func(conn):
            # type: (Connection) -> List[int]
            orders = []
            formatted_limit_date = limit_date.strftime("%Y-%m-%dT%H:%M:%S")
            for row in conn.select(get_orders_to_purge_query.format(formatted_limit_date)):
                order_id = int(row.get_entry("OrderId"))
                orders.append(order_id)

            return orders

        return self.execute_with_connection(func)

    def purge_orders(self, orders):
        def func(conn):
            # type: (Connection) -> None
            orders_list = [str(order) for order in orders]
            query = purge_orders_query.format(",".join(orders_list))
            conn.query(query)

        return self.execute_with_connection(func)

    def save_pick_list_order(
        self,
        order_id,           # type: int
        order_id_count,     # type: str
        jit,                # type: str
        course,             # type: str
        receipt,            # type: str
        status,             # type: bool
    ):
        # type: (...) -> None

        def func(
            conn,   # type: Connection
        ):
            # type: (...) -> None

            date = datetime.now().strftime(
                "%m/%d/%Y %H:%M:%S",
            )

            query = insert_pick_list_query.format(
                order_id,
                order_id_count,
                jit,
                course,
                receipt,
                date,
                status
            )

            conn.query(
                sql=query,
            )

        self.execute_with_connection(func)

    def get_all_pick_list_orders(
        self,
        orders_list,    # type: List[str]
    ):
        # type: (...) -> Dict

        def func(conn):
            # type: (Connection) -> Dict

            orders = [str(order) for order in orders_list]

            query = select_pick_list_orders_query.format(",".join(orders))

            pick_list_orders = {}

            for row in conn.select(query):
                jit = row.get_entry(
                     col="JIT",
                )

                course_name = row.get_entry(
                   col="Course",
                )
                if course_name != '':
                    course_name = " - {}".format(
                        course_name,
                    )

                key = "{}{}".format(
                    jit,
                    course_name,
                )

                pick_list_orders[key] = row.get_entry(
                        col="Id",
                    )

            return pick_list_orders

        return self.execute_with_connection(func)

    def get_data_pick_list_order(
        self,
        pick_list_order_id,     # type: int
    ):
        # type: (...) -> str

        def func(conn):
            # type: (Connection) -> Optional[str]

            query = """SELECT Data FROM PicklistOrders WHERE Id == {}"""

            query = query.format(
                pick_list_order_id,
            )

            for row in conn.select(query):
                return row.get_entry(
                     col="Data",
                )

            return None

        return self.execute_with_connection(func)

insert_query = """
INSERT OR REPLACE INTO ProductionOrders
(OrderId, Object, CreatedAt, Purged)
VALUES
({}, '{}', '{}', {})
"""

select_query = """
SELECT Object FROM ProductionOrders WHERE PURGED = 0 ORDER BY CreatedAt DESC LIMIT 1000
"""

get_orders_to_purge_query = """
SELECT OrderId FROM ProductionOrders WHERE CreatedAt < '{}' ORDER BY CreatedAt ASC LIMIT 1000
"""

purge_orders_query = """
DELETE FROM ProductionOrders WHERE OrderId IN ({})
"""

insert_pick_list_query = """
INSERT OR REPLACE INTO PicklistOrders
(OrderId, OrderIdCount, JIT, Course, Data, CreatedAt, PrintOK)
VALUES
({}, '{}', '{}', '{}', '{}', '{}', {})
"""

select_pick_list_orders_query = """
SELECT Id, JIT, Course FROM PicklistOrders WHERE OrderId in ({}) ORDER BY CreatedAt
"""




from threading import Lock

from msgbus import MBEasyContext
from messagehandler import MessageHandlerBuilder, MessageHandler
from messageprocessor import MessageProcessorMessageHandler, LoggingProcessorCallback, \
    DefaultMessageProcessorExecutorFactory

from customer_services.interactor.update_customer import UpdateCustomerInteractor
from customer_services.service import GetCustomerService
from customer_services.interactor.add_customer import AddCustomerInteractor
from customer_services.interactor.get_customer import GetCustomerInteractor

from externals.processor import UpdateCustomerProcessor
from externals.repository.database import CustomerRepository
from externals.messages import ListenedMessages
from externals.processor import AddCustomerProcessor, GetCustomerProcessor
from externals.util import LoggerUtil, LogManager


database_lock = Lock()
logger = LoggerUtil.get_logger_name(__name__)


class ComponentMessageHandlerBuilder(MessageHandlerBuilder):
    def __init__(self, mb_context, persistence_name):
        # type: (MBEasyContext, str) -> None
        
        super(ComponentMessageHandlerBuilder, self).__init__()

        self.mb_context = mb_context
        self.persistence_name = persistence_name
        
        self.log_managers = None
        self.subject_counter = {}

    def build_singletons(self):
        # type: () -> None

        self.log_managers = LogManager("Subject Counter", self.subject_counter)
        self.log_managers.start()

    def build_message_handler(self):
        # type: () -> MessageHandler
        
        customer_repository = CustomerRepository(self.mb_context, self.persistence_name, database_lock)
        get_customer_service = GetCustomerService(customer_repository)

        add_customer = ListenedMessages.TK_CUSTOMER_ADD_CUSTOMER.value
        get_customer = ListenedMessages.TK_CUSTOMER_GET_CUSTOMER.value
        update_customer = ListenedMessages.TK_CUSTOMER_UPDATE_CUSTOMER.value
        
        add_customer_interactor = AddCustomerInteractor(customer_repository, get_customer_service)
        get_customer_interactor = GetCustomerInteractor(get_customer_service)
        update_customer_interactor = UpdateCustomerInteractor(customer_repository, get_customer_service)
        
        message_processors = {
                add_customer: AddCustomerProcessor(add_customer_interactor),
                get_customer: GetCustomerProcessor(get_customer_interactor),
                update_customer: UpdateCustomerProcessor(update_customer_interactor)
        }

        callbacks = [LoggingProcessorCallback(logger)]
        return MessageProcessorMessageHandler(None,
                                              message_processors,
                                              None,
                                              DefaultMessageProcessorExecutorFactory(callbacks=callbacks),
                                              subject_counter=self.subject_counter,
                                              logger=logger)

    def destroy_message_handler(self, message_handler):
        # type: (MessageHandlerBuilder) -> None
    
        return

    def destroy_singletons(self):
        # type: () -> None
        
        if self.log_managers:
            self.log_managers.stop()



from ._OrderChangerProductionBox import OrderChangerProductionBox


class RemoveDeletedItemsBox(OrderChangerProductionBox):
    def __init__(self, name, sons, logger=None):
        super(RemoveDeletedItemsBox, self).__init__(name, sons, logger)

    def change_order(self, order):
        order.items = self.remove_deleted_items(order.items)
        return order

    @staticmethod
    def remove_deleted_items(items):
        new_items = []
        for item in items:
            if item.qty > 0:
                new_items.append(item)
        return new_items



from _NfceDisablerResponse import NfceDisablerResponse
from _NfceDisablerResponseParser import NfceDisablerResponseParser
from _NfceDisablerParameters import NfceDisablerParameters
from _NfceDisablerRequestBuilder import NfceDisablerRequestBuilder
from _NfceDisabler import NfceDisabler
from _OrderRepository import OrderRepository
from _OrderDisabler import OrderDisabler
from _ProcInutNfeBuilder import ProcInutNfeBuilder
from _ProcInutNfeSaver import ProcInutNfeSaver

__all__ = ["OrderDisabler", "NfceDisablerResponseParser", "NfceDisablerResponse", "NfceDisabler", "NfceDisablerParameters", "NfceDisablerRequestBuilder",
           "OrderRepository", "ProcInutNfeBuilder", "ProcInutNfeSaver"]



import json

from application.model import DefaultBenefitApplierRepository, BenefitAppliers, OperationDescription, OperationStatus
from application.model.configuration import Configurations
from application.repository.pos import DBRepository
from application.repository.pos.utils import find_benefit_applier
from typing import Dict


class ApplyBenefitInteractor(object):

    def __init__(self, configs, db_repository, benefit_appliers_repositories):
        # type: (Configurations, DBRepository, Dict[BenefitAppliers: DefaultBenefitApplierRepository]) -> None

        self.configs = configs
        self.logger = self.configs.logger
        self.db_repository = db_repository
        self.benefit_appliers_repositories = benefit_appliers_repositories

    def apply_benefit(
            self,
            current_pos_id,             # type: int
            originator_pos_id,          # type: int
            voucher_id,                 # type: str
            user_auth_level             # type: bool
    ):
        # type: (...) -> None

        benefit_applier = find_benefit_applier(originator_pos_id, voucher_id)

        message = "Trying to get and lock benefit to voucherId: {}; Current/Originator PosId: {}/{}; BenefitApplier: {}"
        self.logger.info(message.format(voucher_id, current_pos_id, originator_pos_id, benefit_applier))

        benefit = self.benefit_appliers_repositories[benefit_applier].get_benefit_and_lock(voucher_id)
        benefit_id = str(benefit.get("id"))

        try:
            self.logger.info("Benefit successfully obtained. BenefitId: {}".format(benefit_id))
            tender_id = self.configs.benefit_applier_tender_ids[benefit_applier]
            self.benefit_appliers_repositories[benefit_applier].add_benefit(
                    current_pos_id=current_pos_id,
                    benefit_id=benefit_id,
                    benefit=benefit,
                    benefit_applier=benefit_applier,
                    tender_id=tender_id,
                    user_auth_level=user_auth_level
            )

            self.db_repository.insert_or_update_pending_operation(
                benefit_id,
                current_pos_id,
                OperationDescription.APPLIED,
                OperationStatus.DONE,
                benefit_applier,
                0
            )
        except Exception as _:
            self.logger.exception("Error applying benefit. Trying to unlock. BenefitId: {}".format(benefit_id))
            self.benefit_appliers_repositories[benefit_applier].unlock_benefit(benefit_id)
            raise
